import os
import string
import sys
from pathlib import Path
from typing import List

try:
    import pythoncom
except (ModuleNotFoundError, ImportError):
    pythoncom = None

try:
    import win32api
    import win32com.client as win32com_client
except (ModuleNotFoundError, ImportError):
    win32api = None
    win32com_client = None

try:
    import win32net
except (ModuleNotFoundError, ImportError):
    win32net = None

from sickchill import logger, settings


def get_windows_drives_old() -> List[str]:
    if sys.version_info >= (3, 12) and os.name == "nt":
        return os.listdrives()

    if win32api:
        configure_com(True)
        drives = [drive for drive in win32api.GetLogicalDriveStrings().split("\000")][:-1]
        configure_com()
        if drives:
            return drives

    from ctypes import windll

    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()
    for letter in string.ascii_uppercase:
        if bitmask & 1:
            drives.append(f"{letter}:{os.sep}")
        bitmask >>= 1

    return drives


def get_windows_drives() -> List[str]:
    """Return list of detected drives"""

    if win32api is None:
        if os.name == "nt":
            logger.info(_("Unable to get windows shares without pywin32 installed"))

        return get_windows_drives_old()

    configure_com(True)

    # Add Logical Drives
    drives = win32api.GetLogicalDriveStrings().split("\000")[:-1]

    if win32com_client:
        # Add Network Locations (not the same as network shares, these are shortcut files possibly to shares)
        net_shortcuts_location = Path(os.getenv("APPDATA"))
        net_shortcuts_location /= "Microsoft\\Windows\\Network Shortcuts"
        network_shortcuts = []
        for location in net_shortcuts_location.iterdir():
            if location.is_file() and location.suffix == ".lnk":
                network_shortcuts.append(location)
        shell = win32com_client.Dispatch("WScript.Shell")
        for network_shortcut in network_shortcuts:
            shortcut = shell.CreateShortCut(str(network_shortcut))
            drives.append(shortcut.Targetpath)

    configure_com()

    return drives


def get_network_shares():
    network_shares = []
    if win32net is None:
        if os.name == "nt":
            logger.info(_("Unable to get windows shares without pywin32 installed"))
        return network_shares

    configure_com(True)

    servers, count, unknown = win32net.NetServerEnum(None, 100)
    for server in servers:
        # noinspection PyBroadException
        try:
            shares, count, unknown2 = win32net.NetShareEnum(f"\\\\{server['name']}", 0)
            for share in shares:
                network_shares.append(f"\\\\{server['name']}\\{share['netname']}")
        except Exception:
            pass

    configure_com()

    return network_shares


def folders_at_path(path: Path, include_parent: bool = False, include_files: bool = False, file_types: List[str] = None) -> List[dict]:
    """
    Returns a list of dictionaries with the folders contained at the given path.

    Give the empty string as the path to list the contents of the root path
    (under Unix this means "/", on Windows this will be a list of drive letters)

    :param path: to list contents
    :param include_parent: boolean, include parent dir in list as well
    :param include_files: boolean, include files or only directories
    :param file_types: list, file extensions to include, 'images' is an alias for image types
    :return: list of dicts describing folders/files
    """

    # Resolve symlinks and return a sane path
    path = path.resolve()

    # Go up in the path to the containing dir if this is a file
    while not path.is_dir():
        path = path.parent

    if path.parent == path:
        entries = [{"currentPath": ("/", _("My Computer"))[os.name == "nt"]}]
        for name, share in settings.WINDOWS_SHARES.items():
            entries.append({"name": name, "path": f"\\\\{share['server']}\\{share['path']}"})

        if os.name == "nt":
            # noinspection PyBroadException
            try:
                for letter in get_windows_drives():
                    entries.append({"name": letter, "path": letter})

                # TODO: Add a thread to do this on startup and a button in the settings to "find shares", this takes way to long to run.
                # We can store them in WINDOWS_SHARES
                # for share in get_network_shares():
                #     entries.append({"name": share, "path": share})
            except Exception:
                logger.debug("Attempting to get windows drives and shares failed", exc_info=True, stack_info=True)
    else:
        entries = [{"currentPath": str(path.resolve())}]
        if include_parent:
            entries.append({"name": "..", "path": str(path.parent.resolve())})

    hide_list = [
        "boot",
        "bootmgr",
        "cache",
        "config.msi",
        "msocache",
        "recovery",
        "$recycle.bin",
        "recycler",
        "system volume information",
        "temporary internet files",
    ]  # windows specific
    hide_list += [".fseventd", ".spotlight", ".trashes", ".vol", "cachedmessages", "caches", "trash"]  # osx specific
    hide_list += [".git"]
    file_list, dir_list = [], []

    try:
        for entry in path.iterdir():
            if entry.is_file() and not include_files:
                continue

            if str(entry).lower() in hide_list:
                continue

            is_image = False
            is_allowed_type = True

            if entry.is_file() and file_types:
                if "images" in file_types:
                    is_image = entry.suffix in (".jpg", ".jpeg", ".png", ".tiff", ".gif")

                is_allowed_type = entry.suffix in tuple(f".{file_type.strip('.')}" for file_type in file_types) or is_image
                if not is_allowed_type:
                    continue

            file_types = file_types or []

            item_to_add = {"name": entry.name, "path": str(entry.resolve()), "isFile": entry.is_file(), "isImage": is_image, "isAllowed": is_allowed_type}

            if entry.is_file():
                file_list.append(item_to_add)
            else:
                dir_list.append(item_to_add)
    except OSError as error:
        logger.warning(_("Unable to open {path}: {error}").format(path=path, error=error))
        logger.info(_("Unable to open {path}: trying parent directory").format(path=path))
        if path.parent != path:
            return folders_at_path(path.parent, include_parent, include_files, file_types)

    return entries + dir_list + file_list


def configure_com(start: bool = False) -> bool:
    if pythoncom:
        # noinspection PyBroadException
        try:
            if start:
                # noinspection PyUnresolvedReferences
                pythoncom.CoInitializeEx()
            else:
                # noinspection PyUnresolvedReferences
                pythoncom.CoUninitializeEx()

            return True
        except Exception:
            pass

    return False



import logging

from pytest import fixture

from sickchill.views.api import webapi


@fixture(params=set(webapi.function_mapper.values()))
def api_command(request, caplog):
    with caplog.at_level(logging.WARNING, logger="root"):
        caplog.set_level(logging.DEBUG, logger="sickchill")
        caplog.set_level(logging.WARNING, logger="cacheyou")
        caplog.set_level(logging.WARNING, logger="urllib3")
        yield request.param


def test_api_calls(api_command, capsys):
    with capsys.disabled():
        print(api_command._help)

    # TODO: test the api calls, this method will be called once for each api command automatically because of the fixture



import logging
import os
import shutil
import stat
import traceback
from pathlib import Path
from typing import TYPE_CHECKING

import validators
from rarfile import BadRarFile, Error, NeedFirstVolume, PasswordRequired, RarCRCError, RarExecError, RarFile, RarOpenError, RarWrongPassword

from sickchill import logger, settings
from sickchill.helper.common import is_media_file, is_rar_file, is_sync_file, is_torrent_or_nzb_file, remove_extension
from sickchill.helper.exceptions import EpisodePostProcessingFailedException, FailedPostProcessingFailedException

from . import common, db, failedProcessor, helpers, postProcessor

if TYPE_CHECKING:
    from sickchill.oldbeard.name_parser.parser import ParseResult


class ProcessResult(object):
    def __init__(self):
        self.result = True
        self.output = ""
        self.missed_files = []
        self.aggresult = True


def delete_folder(folder, check_empty=True):
    """
    Removes a folder from the filesystem

    param folder: Path to folder to remove
    param check_empty: Boolean, check if the folder is empty before removing it, defaults to True
    return: True on success, False on failure
    """

    folder = Path(folder).resolve()
    # check if it's a folder
    if not folder.is_dir():
        return False

    # check if it isn't TV_DOWNLOAD_DIR
    if settings.TV_DOWNLOAD_DIR and str(Path(folder).resolve()) == str(Path(settings.TV_DOWNLOAD_DIR).resolve()):
        return False

    # check if it's empty folder when wanted to be checked
    if check_empty:
        found_files = [file for file in folder.iterdir()]
        if found_files:
            logger.info(f"Not deleting folder {folder} found the following files: {found_files}")
            return False

        try:
            logger.info(f"Deleting folder (if it's empty): {folder}")
            folder.rmdir()
        except (OSError, IOError) as error:
            logger.warning(f"Warning: unable to delete folder: {folder}: {error}")
            return False
    else:
        try:
            logger.info(f"Deleting folder: {folder}")
            shutil.rmtree(folder)
        except (OSError, IOError) as error:
            logger.warning(f"Warning: unable to delete folder: {folder}: {error}")
            return False

    return True


def delete_files(process_path, unwanted_files, result, force=False):
    """
    Remove files from filesystem

    param process_path: path to process
    param unwanted_files: files we do not want
    param result: Processor results
    param force: Boolean, force deletion, defaults to false
    """
    if not result.result and force:
        result.output += log_helper("Forcing deletion of files, even though last result was not success", logger.DEBUG)
    elif not result.result:
        return

    process_path = Path(process_path)
    # Delete all file not needed
    for current_file in unwanted_files:
        file_path = process_path / current_file
        if not file_path.is_file():
            continue  # Prevent error when a notwantedfiles is an associated files

        result.output += log_helper(f"Deleting file: {current_file}", logger.DEBUG)

        # check first the read-only attribute
        file_attribute = file_path.stat()[0]
        if not file_attribute & stat.S_IWRITE:
            # File is read-only, so make it writeable
            result.output += log_helper(f"Changing ReadOnly Flag for file: {current_file}", logger.DEBUG)
            try:
                file_path.chmod(stat.S_IWRITE)
            except OSError as error:
                result.output += log_helper(f"Cannot change permissions of {current_file}: {error}", logger.DEBUG)
        try:
            file_path.unlink(True)
        except OSError as error:
            result.output += log_helper(f"Unable to delete file {current_file}: {error}", logger.DEBUG)


def log_helper(message, level=logging.INFO):
    logger.log(level, message)
    return message + "\n"


def process_dir(process_path, release_name=None, process_method=None, force=False, is_priority=None, delete_on=False, failed=False, mode="auto"):
    """
    Scans through the files in process_path and processes whatever media files it finds

    param process_path: The folder name to look in
    param release_name: The NZB/Torrent name which resulted in this folder being downloaded
    param process_method: processing method, copy/move/symlink/link
    param force: True to process previously processed files
    param is_priority: whether to replace the file even if it exists at higher quality
    param delete_on: delete files and folders after they are processed (always happens with move and auto combination)
    param failed: Boolean for whether the download failed
    param mode: Type of postprocessing auto or manual
    """
    result = ProcessResult()
    try:
        # if they passed us a real dir then assume it's the one we want
        if os.path.isdir(process_path):
            process_path = os.path.realpath(process_path)
            result.output += log_helper(f"Processing in folder {process_path}", logger.DEBUG)

        # if the client and SickChill are not on the same machine translate the directory into a network directory
        elif all(
            [
                settings.TV_DOWNLOAD_DIR,
                Path(settings.TV_DOWNLOAD_DIR).is_dir(),
                str(Path(process_path).resolve()) == str(Path(settings.TV_DOWNLOAD_DIR).resolve()),
            ]
        ):
            process_path = os.path.join(settings.TV_DOWNLOAD_DIR, os.path.abspath(process_path).split(os.path.sep)[-1])
            result.output += log_helper(f"Trying to use folder: {process_path} ", logger.DEBUG)

        # if we didn't find a real dir then quit
        if not Path(process_path).is_dir():
            result.output += log_helper(
                "Unable to figure out what folder to process. "
                "If your downloader and SickChill aren't on the same PC "
                "make sure you fill out your TV download dir in the config.",
                logger.DEBUG,
            )
            return result.output

        process_method = process_method or settings.PROCESS_METHOD

        directories_from_rars = set()

        # If we have a release name (probably from nzbToMedia), and it is a rar/video, only process that file
        if release_name and validators.url(release_name) is True:
            result.output += log_helper(_("Processing {release_name}").format(release_name=release_name))
            generator_to_use = [("", [], [release_name])]
        elif release_name and (is_media_file(release_name) or is_rar_file(release_name)):
            result.output += log_helper(_("Processing {release_name}").format(release_name=release_name))
            generator_to_use = [(process_path, [], [release_name])]
        else:
            result.output += log_helper(_("Processing {process_path}").format(process_path=process_path))
            generator_to_use = os.walk(process_path, followlinks=settings.PROCESSOR_FOLLOW_SYMLINKS)

        rar_files = []

        for current_directory, directory_names, filenames in generator_to_use:
            result.result = True

            if current_directory:
                filenames = [f for f in filenames if not is_torrent_or_nzb_file(f)]
                rar_files = [x for x in filenames if is_rar_file(os.path.join(current_directory, x))]
                if rar_files:
                    extracted_directories = unrar(current_directory, rar_files, force, result)
                    if extracted_directories:
                        for extracted_directory in extracted_directories:
                            if extracted_directory.split(current_directory)[-1] not in directory_names:
                                result.output += log_helper(
                                    _("Adding extracted directory to the list of directories to process: {extracted_directory}").format(
                                        extracted_directory=extracted_directory
                                    ),
                                    logger.DEBUG,
                                )
                                directories_from_rars.add(extracted_directory)

            if not validate_dir(current_directory, release_name, failed, result):
                continue

            video_files = list(filter(is_media_file, filenames))
            if video_files:
                process_media(current_directory, video_files, release_name, process_method, force, is_priority, result)
            else:
                result.result = False

            # Delete all file not needed and avoid deleting files if Manual PostProcessing
            if not (process_method == "move" and result.result) or (mode == "manual" and not delete_on):
                continue

            # noinspection PyTypeChecker
            unwanted_files = [x for x in filenames if x in video_files + rar_files]
            if unwanted_files:
                result.output += log_helper(_("Found unwanted files: {unwanted_files}").format(unwanted_files=unwanted_files), logger.DEBUG)

            delete_folder(os.path.join(current_directory, "@eaDir"), False)
            delete_files(current_directory, unwanted_files, result)
            if delete_folder(current_directory, check_empty=not delete_on):
                result.output += log_helper(_("Deleted folder: {current_directory}").format(current_directory=current_directory), logger.DEBUG)

        # For processing extracted rars, only allow methods 'move' and 'copy'.
        # On different methods fall back to 'move'.
        method_fallback = ("move", process_method)[process_method in ("move", "copy")]

        for directory_from_rar in directories_from_rars:
            process_dir(
                process_path=directory_from_rar,
                release_name=os.path.basename(directory_from_rar),
                process_method=method_fallback,
                force=force,
                is_priority=is_priority,
                delete_on=settings.DELRARCONTENTS or delete_on or method_fallback == "move",
                failed=failed,
                mode=mode,
            )

            # Delete rar file only if the extracted dir was successfully processed
            if mode == "auto" and method_fallback == "move" or mode == "manual" and delete_on:
                this_rar = [rar_file for rar_file in rar_files if Path(directory_from_rar).name == Path(rar_file).stem]
                delete_files(current_directory, this_rar, result)  # Deletes only if result.result == True

            delete_folder(directory_from_rar, settings.DELRARCONTENTS)

        result.output += log_helper((_("Processing Failed"), _("Successfully processed"))[result.aggresult], (logger.WARNING, logger.INFO)[result.aggresult])
        if result.missed_files:
            result.output += log_helper(_("Some items were not processed."))
            for missed_file in result.missed_files:
                result.output += log_helper(missed_file)

        return result.output
    except Exception:
        logger.debug(traceback.format_exc())
        return result.output


def validate_dir(process_path, release_name, failed, result):
    """
    Check if directory is valid for processing

    param process_path: Directory to check
    param release_name: Original NZB/Torrent name
    param failed: Previously failed objects
    param result: Previous results
    returns True if dir is valid for processing, False if not
    """

    result.output += log_helper("Processing folder " + process_path, logger.DEBUG)
    upper_name = os.path.basename(process_path).upper()
    if upper_name.startswith("_FAILED_") or upper_name.endswith("_FAILED_") or (os.sep + "_FAILED_") in upper_name or ("_FAILED_" + os.sep) in upper_name:
        result.output += log_helper(_("The directory name indicates it failed to extract."), logger.DEBUG)
        failed = True
    elif (
        upper_name.startswith("_UNDERSIZED_")
        or upper_name.endswith("_UNDERSIZED_")
        or (os.sep + "_UNDERSIZED_") in upper_name
        or ("_UNDERSIZED_" + os.sep) in upper_name
    ):
        result.output += log_helper(_("The directory name indicates that it was previously rejected for being undersized."), logger.DEBUG)
        failed = True
    elif upper_name.startswith("_UNPACK") or upper_name.endswith("_UNPACK") or (os.sep + "_UNPACK") in upper_name or ("_UNPACK" + os.sep) in upper_name:
        result.output += log_helper(_("The directory name indicates that this release is in the process of being unpacked."), logger.DEBUG)
        result.missed_files.append(f"{process_path} : Being unpacked")
        return False

    if failed:
        process_failed(process_path, release_name, result)
        result.missed_files.append(f"{process_path} : Failed download")
        return False

    if (
        settings.TV_DOWNLOAD_DIR
        and str(Path(process_path).resolve()) != str(Path(settings.TV_DOWNLOAD_DIR).resolve())
        and helpers.is_hidden_folder(process_path)
    ):
        result.output += log_helper(f"Ignoring hidden folder: {process_path}", logger.DEBUG)
        if not process_path.endswith("@eaDir"):
            result.missed_files.append(f"{process_path} : Hidden folder")
        return False

    # make sure the dir isn't inside a show dir
    main_db_con = db.DBConnection()
    sql_results = main_db_con.select("SELECT location FROM tv_shows")

    for sqlShow in sql_results:
        if (
            process_path.lower().startswith(os.path.realpath(sqlShow["location"]).lower() + os.sep)
            or process_path.lower() == os.path.realpath(sqlShow["location"]).lower()
        ):
            result.output += log_helper("Cannot process an episode that's already been moved to its show dir, skipping " + process_path, logger.WARNING)
            return False

    for current_directory, directory_names, filenames in os.walk(process_path, topdown=False, followlinks=settings.PROCESSOR_FOLLOW_SYMLINKS):
        sync_files = list(filter(is_sync_file, filenames))
        if sync_files and settings.POSTPONE_IF_SYNC_FILES:
            result.output += log_helper(f"Found temporary sync files: {sync_files} in path: {os.path.join(process_path, sync_files[0])}")
            result.output += log_helper(f"Skipping post processing for folder: {process_path}")
            result.missed_files.append(f"{os.path.join(process_path, sync_files[0])} : Sync files found")
            continue

        found_files = list(filter(is_media_file, filenames))
        if settings.UNPACK == settings.UNPACK_PROCESS_CONTENTS:
            found_files += list(filter(is_rar_file, filenames))

        for found_file in found_files:
            if current_directory != settings.TV_DOWNLOAD_DIR and found_files:
                # pass 'current directory/filename' as one string to NameParser
                found_file = f"{os.path.basename(current_directory)}/{found_file}"

            if postProcessor.guessit_findit(found_file):
                return True

    result.output += log_helper(f"{process_path} : No processable items found in folder", logger.DEBUG)
    return False


def unrar(path, rar_files, force, result):
    """
    Extracts RAR files

    param path: Path to look for files in
    param rar_files: Names of RAR files
    param force: process currently processing items
    param result: Previous results
    returns List of unpacked file names
    """

    unpacked_dirs = []

    if settings.UNPACK == settings.UNPACK_PROCESS_CONTENTS and rar_files:
        result.output += log_helper(f"Packed Releases detected: {rar_files}", logger.DEBUG)
        for archive in rar_files:
            failure = None
            rar_handle = None
            try:
                archive_path = os.path.join(path, archive)
                if already_processed(path, archive, force, result):
                    result.output += log_helper(f"Archive file already post-processed, extraction skipped: {archive_path}", logger.DEBUG)
                    continue

                if not is_rar_file(archive_path):
                    continue

                result.output += log_helper(f"Checking if archive is valid and contains a video: {archive_path}", logger.DEBUG)
                rar_handle = RarFile(archive_path)
                if rar_handle.needs_password():
                    # TODO: Add support in settings for a list of passwords to try here with rar_handle.set_password(x)
                    result.output += log_helper(f"Archive needs a password, skipping: {archive_path}")
                    continue

                rar_handle.testrar()

                # If there are no video files in the rar, don't extract it
                rar_media_files = list(filter(is_media_file, rar_handle.namelist()))
                if not rar_media_files:
                    continue

                rar_release_name = Path(archive).stem

                # Choose the directory we'll unpack to:
                if settings.UNPACK_DIR and os.path.isdir(settings.UNPACK_DIR):  # verify that the unpacked dir exists
                    unpack_base_dir = settings.UNPACK_DIR
                else:
                    unpack_base_dir = path
                    if settings.UNPACK_DIR:  # Let user know if we can't unpack there
                        result.output += log_helper(f"Unpack directory cannot be verified. Using {path}", logger.DEBUG)

                # Fix up the list for checking if already processed
                rar_media_files = [os.path.join(unpack_base_dir, rar_release_name, rar_media_file) for rar_media_file in rar_media_files]

                skip_rar = False
                for rar_media_file in rar_media_files:
                    check_path, check_file = os.path.split(rar_media_file)
                    if already_processed(check_path, check_file, force, result):
                        result.output += log_helper(f"Archive file already post-processed, extraction skipped: {rar_media_file}", logger.DEBUG)
                        skip_rar = True
                        break

                if skip_rar:
                    continue

                rar_extract_path = os.path.join(unpack_base_dir, rar_release_name)
                result.output += log_helper(f"Unpacking archive: {archive}", logger.DEBUG)
                rar_handle.extractall(path=rar_extract_path)
                unpacked_dirs.append(rar_extract_path)

            except RarCRCError:
                failure = ("Archive Broken", "Unpacking failed because of a CRC error")
            except RarWrongPassword:
                failure = ("Incorrect RAR Password", "Unpacking failed because of an Incorrect Rar Password")
            except PasswordRequired:
                failure = ("Rar is password protected", "Unpacking failed because it needs a password")
            except RarOpenError:
                failure = (
                    "Rar Open Error, check the parent folder and destination file permissions.",
                    "Unpacking failed with a File Open Error (file permissions?)",
                )
            except RarExecError:
                failure = ("Invalid Rar Archive Usage", "Unpacking Failed with Invalid Rar Archive Usage. Is unrar installed and on the system PATH?")
            except BadRarFile:
                failure = ("Invalid Rar Archive", "Unpacking Failed with an Invalid Rar Archive Error")
            except NeedFirstVolume:
                continue
            except (Exception, Error) as error:
                failure = (error, "Unpacking failed")
            finally:
                if rar_handle:
                    del rar_handle

            if failure:
                result.output += log_helper(f"Failed to extract the archive {archive}: {failure[0]}", logger.WARNING)
                result.missed_files.append(f"{archive} : Unpacking failed: {failure[1]}")
                result.result = False
                continue

    return unpacked_dirs


def already_processed(process_path, video_file, force, result):
    """
    Check if we already post processed a file

    param process_path: Directory a file resides in
    param video_file: File name
    param force: Force checking when already checking (currently unused)
    param result: True if file is already postprocessed, False if not
    :return:
    """
    if force:
        return False

    # Avoid processing the same dir again if we use a process method <> move
    main_db_con = db.DBConnection()
    sql_result = main_db_con.select("SELECT release_name FROM tv_episodes WHERE release_name IN (?, ?) LIMIT 1", [process_path, remove_extension(video_file)])
    if sql_result:
        # result.output += log_helper("You're trying to post process a dir that's already been processed, skipping", logger.DEBUG)
        return True

    # Needed if we have downloaded the same episode @ different quality
    # But we need to make sure we check the history of the episode we're going to PP, and not others
    # if it fails to find any info (because we're doing an unparsable folder (like the TV root dir) it will throw an exception, which we want to ignore
    parse_result: "ParseResult" = postProcessor.guessit_findit(process_path)

    search_sql = "SELECT tv_episodes.indexerid, history.resource FROM tv_episodes INNER JOIN history ON history.showid=tv_episodes.showid"  # This part is always the same
    search_sql += " WHERE history.season=tv_episodes.season AND history.episode=tv_episodes.episode"

    # If we find a showid, a season number, and one or more episode numbers than we need to use those in the query
    if parse_result:
        if parse_result.show.indexerid:
            search_sql += f" AND tv_episodes.showid={parse_result.show.indexerid}"
        if parse_result.season_number is not None and parse_result.episode_numbers:
            search_sql += f" AND tv_episodes.season={parse_result.season_number} AND tv_episodes.episode={parse_result.episode_numbers[0]}"
        elif parse_result.ab_episode_numbers:
            search_sql += f" AND tv_episodes.showid={parse_result.show.indexerid} AND tv_episodes.absolute_number={parse_result.ab_episode_numbers[0]}"

    search_sql += " AND tv_episodes.status IN (" + ",".join([str(x) for x in common.Quality.DOWNLOADED + common.Quality.ARCHIVED]) + ")"
    search_sql += " AND history.resource LIKE ? LIMIT 1"
    sql_result = main_db_con.select(search_sql, ["%" + video_file])
    if sql_result:
        result.output += log_helper("You're trying to post process a video that's already been processed, skipping", logger.DEBUG)
        return True

    return False


def process_media(process_path, video_files, release_name, process_method, force, is_priority, result):
    """
    Postprocess mediafiles

    param process_path: Path to process in
    param video_files: Filenames to look for and postprocess
    param release_name: Name of NZB/Torrent file related
    param process_method: auto/manual
    param force: Postprocess currently postprocessing file
    param is_priority: Boolean, is this a priority download
    param result: Previous results
    """

    processor = None
    for cur_video_file in video_files:
        cur_video_file_path = os.path.join(process_path, cur_video_file)

        if already_processed(process_path, cur_video_file, force, result):
            result.output += log_helper(f"Skipping already processed file: {cur_video_file}", logger.DEBUG)
            continue

        try:
            processor = postProcessor.PostProcessor(cur_video_file_path, release_name, process_method, is_priority)
            result.result = processor.process()
            process_fail_message = ""
        except EpisodePostProcessingFailedException as error:
            result.result = False
            process_fail_message = error

        if processor:
            result.output += processor.log

        if result.result:
            result.output += log_helper(f"Processing succeeded for {cur_video_file_path}")
        else:
            result.output += log_helper(f"Processing failed for {cur_video_file_path}: {process_fail_message}", logger.WARNING)
            result.missed_files.append(f"{cur_video_file_path} : Processing failed: {process_fail_message}")
            result.aggresult = False


def process_failed(process_path, release_name, result):
    """Process a download that did not complete correctly"""

    if not settings.USE_FAILED_DOWNLOADS:
        return

    processor = None

    try:
        processor = failedProcessor.FailedProcessor(process_path, release_name)
        result.result = processor.process()
        process_fail_message = ""
    except FailedPostProcessingFailedException as error:
        result.result = False
        process_fail_message = error

    if processor:
        result.output += processor.log

    if settings.DELETE_FAILED and result.result:
        if delete_folder(process_path, check_empty=False):
            result.output += log_helper(f"Deleted folder: {process_path}", logger.DEBUG)

    if result.result:
        result.output += log_helper(f"Failed Download Processing succeeded: ({release_name}, {process_path})")
    else:
        result.output += log_helper(f"Failed Download Processing failed: ({release_name}, {process_path}): {process_fail_message}", logger.WARNING)



import datetime
import os
import re
from xml.etree import ElementTree

import sickchill
from sickchill import logger, settings
from sickchill.helper.common import dateFormat, replace_extension
from sickchill.oldbeard import helpers

from . import generic


class MediaBrowserMetadata(generic.GenericMetadata):
    """
    Metadata generation class for Media Browser 2.x/3.x - Standard Mode.

    The following file structure is used:

    show_root/series.xml                       (show metadata)
    show_root/folder.jpg                       (poster)
    show_root/backdrop.jpg                     (fanart)
    show_root/Season ##/folder.jpg             (season thumb)
    show_root/Season ##/filename.ext           (*)
    show_root/Season ##/metadata/filename.xml  (episode metadata)
    show_root/Season ##/metadata/filename.jpg  (episode thumb)
    """

    def __init__(
        self,
        show_metadata=False,
        episode_metadata=False,
        fanart=False,
        poster=False,
        banner=False,
        episode_thumbnails=False,
        season_posters=False,
        season_banners=False,
        season_all_poster=False,
        season_all_banner=False,
    ):
        super().__init__(
            show_metadata, episode_metadata, fanart, poster, banner, episode_thumbnails, season_posters, season_banners, season_all_poster, season_all_banner
        )

        self.name = "MediaBrowser"

        self._ep_nfo_extension = "xml"
        self._show_metadata_filename = "series.xml"

        self.fanart_name = "backdrop.jpg"
        self.poster_name = "folder.jpg"

        # web-ui metadata template
        self.eg_show_metadata = "series.xml"
        self.eg_episode_metadata = "Season##\\metadata\\<i>filename</i>.xml"
        self.eg_fanart = "backdrop.jpg"
        self.eg_poster = "folder.jpg"
        self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\metadata\\<i>filename</i>.jpg"
        self.eg_season_posters = "Season##\\folder.jpg"
        self.eg_season_banners = "Season##\\banner.jpg"
        self.eg_season_all_poster = "<i>not supported</i>"
        self.eg_season_all_banner = "<i>not supported</i>"

    # Override with empty methods for unsupported features
    def retrieveShowMetadata(self, folder):
        # while show metadata is generated, it is not supported for our lookup
        return None, None, None

    def create_season_all_poster(self, show_obj):
        pass

    def create_season_all_banner(self, show_obj):
        pass

    def get_episode_file_path(self, episode_object):
        """
        Returns a full show dir/metadata/episode.xml path for MediaBrowser
        episode metadata files

        episode_object: a TVEpisode object to get the path for
        """

        if os.path.isfile(episode_object.location):
            xml_filename = replace_extension(os.path.basename(episode_object.location), self._ep_nfo_extension)
            metadata_dir_name = os.path.join(os.path.dirname(episode_object.location), "metadata")
            xml_file_path = os.path.join(metadata_dir_name, xml_filename)
        else:
            logger.debug(f"[{self.name} META] Episode location doesn't exist: {episode_object.location}")
            return ""

        return xml_file_path

    @staticmethod
    def get_episode_thumb_path(episode_object):
        """
        Returns a full show dir/metadata/episode.jpg path for MediaBrowser
        episode thumbs.

        episode_object: a TVEpisode object to get the path from
        """

        if os.path.isfile(episode_object.location):
            tbn_filename = replace_extension(os.path.basename(episode_object.location), "jpg")
            metadata_dir_name = os.path.join(os.path.dirname(episode_object.location), "metadata")
            tbn_file_path = os.path.join(metadata_dir_name, tbn_filename)
        else:
            return None

        return tbn_file_path

    @staticmethod
    def get_season_poster_path(show_obj, season):
        """
        Season thumbs for MediaBrowser go in Show Dir/Season X/folder.jpg

        If no season folder exists, None is returned
        """

        dir_list = [x for x in os.listdir(show_obj.location) if os.path.isdir(os.path.join(show_obj.location, x))]

        season_dir_regex = r"^Season\s+(\d+)$"

        season_dir = None

        for cur_dir in dir_list:
            # MediaBrowser 1.x only supports 'Specials'
            # MediaBrowser 2.x looks to only support 'Season 0'
            # MediaBrowser 3.x looks to mimic KODI/Plex support
            if season == 0 and cur_dir == "Specials":
                season_dir = cur_dir
                break

            match = re.match(season_dir_regex, cur_dir, re.I)
            if not match:
                continue

            cur_season = int(match.group(1))

            if cur_season == season:
                season_dir = cur_dir
                break

        if not season_dir:
            logger.debug(f"Unable to find a season dir for season {season}")
            return None

        logger.debug(f"Using {season_dir}/folder.jpg as season dir for season {season}")

        return os.path.join(show_obj.location, season_dir, "folder.jpg")

    @staticmethod
    def get_season_banner_path(show_obj, season):
        """
        Season thumbs for MediaBrowser go in Show Dir/Season X/banner.jpg

        If no season folder exists, None is returned
        """

        dir_list = [x for x in os.listdir(show_obj.location) if os.path.isdir(os.path.join(show_obj.location, x))]

        season_dir_regex = r"^Season\s+(\d+)$"

        season_dir = None

        for cur_dir in dir_list:
            # MediaBrowser 1.x only supports 'Specials'
            # MediaBrowser 2.x looks to only support 'Season 0'
            # MediaBrowser 3.x looks to mimic KODI/Plex support
            if season == 0 and cur_dir == "Specials":
                season_dir = cur_dir
                break

            match = re.match(season_dir_regex, cur_dir, re.I)
            if not match:
                continue

            cur_season = int(match.group(1))

            if cur_season == season:
                season_dir = cur_dir
                break

        if not season_dir:
            logger.debug(f"Unable to find a season dir for season {season}")
            return None

        logger.debug(f"Using {season_dir}/banner.jpg as season dir for season {season}")

        return os.path.join(show_obj.location, season_dir, "banner.jpg")

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for a MediaBrowser-style series.xml
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        indexer_show = sickchill.indexer.series(show_obj)
        if not indexer_show:
            logger.exception("Unable to find show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.indexer_name))
            return False

        # check for title and id
        if not (getattr(indexer_show, "seriesName", None) and getattr(indexer_show, "id", None)):
            logger.info("Incomplete info for show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.indexer_name))
            return False

        tv_node = ElementTree.Element("Series")

        data = getattr(indexer_show, "id", None)
        if data:
            indexerid_element = ElementTree.SubElement(tv_node, "id")
            indexerid_element.text = str(data)

        data = getattr(indexer_show, "seriesName", None)
        if data:
            series_name_element = ElementTree.SubElement(tv_node, "SeriesName")
            series_name_element.text = data

        data = getattr(indexer_show, "status", None)
        if data:
            status_element = ElementTree.SubElement(tv_node, "Status")
            status_element.text = data

        data = getattr(indexer_show, "network", None)
        if data:
            network_element = ElementTree.SubElement(tv_node, "Network")
            network_element.text = data

            studios_element = ElementTree.SubElement(tv_node, "Studios")
            studio_element = ElementTree.SubElement(studios_element, "Studio")
            studio_element.text = data

        data = getattr(indexer_show, "airsTime", None)
        if data:
            airs_time_element = ElementTree.SubElement(tv_node, "Airs_Time")
            airs_time_element.text = data

        data = getattr(indexer_show, "airsDayOfWeek", None)
        if data:
            airs_day_element = ElementTree.SubElement(tv_node, "Airs_DayOfWeek")
            airs_day_element.text = data

        data = getattr(indexer_show, "firstAired", None)
        if data:
            first_aired_element = ElementTree.SubElement(tv_node, "FirstAired")
            first_aired_element.text = data
            premier_date_element = ElementTree.SubElement(tv_node, "PremiereDate")
            premier_date_element.text = data
            try:
                year_text = str(datetime.datetime.strptime(data, dateFormat).year)
                if year_text:
                    production_year_element = ElementTree.SubElement(tv_node, "ProductionYear")
                    production_year_element.text = year_text
            except Exception:
                pass

        data = getattr(indexer_show, "rating", None)
        if data:
            content_rating_element = ElementTree.SubElement(tv_node, "ContentRating")
            content_rating_element.text = data

            mpaa_rating_element = ElementTree.SubElement(tv_node, "MPAARating")
            mpaa_rating_element.text = data

            certification_element = ElementTree.SubElement(tv_node, "certification")
            certification_element.text = data

            rating_element = ElementTree.SubElement(tv_node, "Rating")
            rating_element.text = data

        type_element = ElementTree.SubElement(tv_node, "Type")
        type_element.text = "Series"

        data = getattr(indexer_show, "overview", None)
        if data:
            overview_element = ElementTree.SubElement(tv_node, "Overview")
            overview_element.text = data

        data = getattr(indexer_show, "runtime", None)
        if data:
            running_time_element = ElementTree.SubElement(tv_node, "RunningTime")
            running_time_element.text = data

            runtime_element = ElementTree.SubElement(tv_node, "Runtime")
            runtime_element.text = data

        data = getattr(indexer_show, "imdbId", None)
        if data:
            imdb_id_element = ElementTree.SubElement(tv_node, "IMDB_ID")
            imdb_id_element.text = data

            imdb_id_element = ElementTree.SubElement(tv_node, "IMDB")
            imdb_id_element.text = data

            imdb_id_element = ElementTree.SubElement(tv_node, "IMDbId")
            imdb_id_element.text = data

        data = getattr(indexer_show, "zap2itId", None)
        if data:
            zap2itid_element = ElementTree.SubElement(tv_node, "Zap2ItId")
            zap2itid_element.text = data

        if getattr(indexer_show, "genre", []) and isinstance(indexer_show.genre, list):
            genres_element = ElementTree.SubElement(tv_node, "Genres")
            for genre in indexer_show.genre:
                if genre.strip():
                    genre_element = ElementTree.SubElement(genres_element, "Genre")
                    genre_element.text = genre.strip()

            genre_element = ElementTree.SubElement(tv_node, "Genre")
            genre_element.text = "|".join(indexer_show.genre)

        helpers.indentXML(tv_node)

        return ElementTree.ElementTree(tv_node)

    def _ep_data(self, episode_object):
        """
        Creates an elementTree XML structure for a MediaBrowser style episode.xml
        and returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        episodes_to_write = [episode_object] + episode_object.related_episodes

        persons_dict = {"Director": [], "GuestStar": [], "Writer": []}

        indexer_show = episode_object.idxr.series_from_episode(episode_object)
        if not indexer_show:
            logger.info("Unable to connect to {} while creating meta files - skipping".format(episode_object.idxr.name))
            return False

        root_node = ElementTree.Element("Item")

        # write an MediaBrowser XML containing info for all matching episodes
        for current_episode in episodes_to_write:
            indexer_episode = current_episode.idxr.episode(current_episode)
            if not indexer_episode:
                continue

            if current_episode == episode_object:
                # root (or single) episode

                # default to today's date for specials if firstAired is not set
                if episode_object.airdate != datetime.date.min and indexer_episode.get("firstAired"):
                    indexer_episode["firstAired"] = str(episode_object.airdate)

                if episode_object.season == 0 and not indexer_episode.get("firstAired"):
                    if episode_object.show and episode_object.show.startyear:
                        indexer_episode["firstAired"] = str(datetime.date.min.replace(year=episode_object.show.startyear))
                    else:
                        indexer_episode["firstAired"] = str(datetime.date.today())

                if not (indexer_episode.get("episodeName") and indexer_episode.get("firstAired")):
                    return None

                episode = root_node

                if current_episode.name:
                    episode_name_element = ElementTree.SubElement(episode, "EpisodeName")
                    episode_name_element.text = current_episode.name

                episode_number_element = ElementTree.SubElement(episode, "EpisodeNumber")
                episode_number_element.text = str(episode_object.episode)

                if episode_object.related_episodes:
                    try:
                        max_episode = max(e.episode for e in episode_object.related_episodes if e)
                    except (AttributeError, TypeError):
                        max_episode = current_episode.episode

                    episode_number_end_element = ElementTree.SubElement(episode, "EpisodeNumberEnd")
                    episode_number_end_element.text = str(max_episode)

                season_number_element = ElementTree.SubElement(episode, "SeasonNumber")
                season_number_element.text = str(current_episode.season)

                if indexer_episode.get("absoluteNumber") and not episode_object.related_episodes:
                    absolute_number_element = ElementTree.SubElement(episode, "absolute_number")
                    absolute_number_element.text = str(indexer_episode["absoluteNumber"])

                if current_episode.airdate != datetime.date.min:
                    first_aired_element = ElementTree.SubElement(episode, "FirstAired")
                    first_aired_element.text = str(current_episode.airdate)

                type_element = ElementTree.SubElement(episode, "Type")
                type_element.text = "Episode"

                if current_episode.description:
                    overview_element = ElementTree.SubElement(episode, "Overview")
                    overview_element.text = current_episode.description
                elif indexer_episode.get("overview"):
                    overview_element = ElementTree.SubElement(episode, "Overview")
                    overview_element.text = indexer_episode["overview"]

                if not episode_object.related_episodes:
                    if indexer_episode.get("rating"):
                        rating_element = ElementTree.SubElement(episode, "Rating")
                        rating_element.text = indexer_episode["siteRating"]

                    if indexer_episode.get("imdbId"):
                        imdb_id_element = ElementTree.SubElement(episode, "IMDB_ID")
                        imdb_id_element.text = indexer_episode["imdbId"]

                        imdb_element = ElementTree.SubElement(episode, "IMDB")
                        imdb_element.text = indexer_episode["imdbId"]

                        imdbid_element = ElementTree.SubElement(episode, "IMDbId")
                        imdbid_element.text = indexer_episode["imdbId"]

                if indexer_episode.get("id"):
                    id_element = ElementTree.SubElement(episode, "id")
                    id_element.text = str(indexer_episode["id"])

                persons_element = ElementTree.SubElement(episode, "Persons")

                for actor in episode_object.idxr.actors(indexer_show):
                    if not ("name" in actor and actor["name"].strip()):
                        continue

                    actor_element = ElementTree.SubElement(persons_element, "Person")

                    actor_name_element = ElementTree.SubElement(actor_element, "Name")
                    actor_name_element.text = actor["name"].strip()

                    actor_type_element = ElementTree.SubElement(actor_element, "Type")
                    actor_type_element.text = "Actor"

                    if "role" in actor and actor["role"].strip():
                        actor_role_element = ElementTree.SubElement(actor_element, "Role")
                        actor_role_element.text = actor["role"].strip()

                if indexer_episode.get("language"):
                    language = indexer_episode.get("language")["overview"]
                else:
                    language = settings.INDEXER_DEFAULT_LANGUAGE

                language_element = ElementTree.SubElement(episode, "Language")
                language_element.text = language

                if indexer_episode.get("filename"):
                    thumb_element = ElementTree.SubElement(episode, "filename")
                    thumb_element.text = current_episode.idxr.complete_image_url(indexer_episode["filename"])

            else:
                # append data from (if any) related episodes
                if current_episode.name:
                    if not episode_name_element.text:
                        episode_name_element.text = current_episode.name
                    else:
                        episode_name_element.text = episode_name_element.text + ", " + current_episode.name

                if current_episode.description:
                    if not overview_element.text:
                        overview_element.text = current_episode.description
                    else:
                        overview_element.text = overview_element.text + "\r" + current_episode.description

            # collect all directors, guest stars and writers
            if indexer_episode.get("directors"):
                persons_dict["Director"] += indexer_episode["directors"]
            if indexer_episode.get("guestStars"):
                persons_dict["GuestStar"] += indexer_episode["guestStars"]
            if indexer_episode.get("writers"):
                persons_dict["Writer"] += indexer_episode["writers"]

        # fill in Persons section with collected directors, guest starts and writers
        for person_type, names in persons_dict.items():
            # remove doubles
            names = list(set(names))
            for cur_name in names:
                person_element = ElementTree.SubElement(persons_element, "Person")
                person_name_element = ElementTree.SubElement(person_element, "Name")
                person_name_element.text = cur_name
                person_type_element = ElementTree.SubElement(person_element, "Type")
                person_type_element.text = person_type

        helpers.indentXML(root_node)

        return ElementTree.ElementTree(root_node)


# present a standard "interface" from the module
metadata_class = MediaBrowserMetadata



import json

import requests
from requests.structures import CaseInsensitiveDict

from sickchill import logger, settings
from sickchill.oldbeard import common


class Notifier(object):
    MATTERMOST_WEBHOOK_URL = ""
    MATTERMOST_ICON_URL = "https://github.com/SickChill/SickChill/raw/master/sickchill/gui/slick/images/sickchill-sc.png"

    def notify_snatch(self, ep_name):
        if settings.MATTERMOST_NOTIFY_SNATCH:
            self._notify_mattermost(common.notifyStrings[common.NOTIFY_SNATCH] + ": " + ep_name)

    def notify_download(self, ep_name):
        if settings.MATTERMOST_NOTIFY_DOWNLOAD:
            self._notify_mattermost(common.notifyStrings[common.NOTIFY_DOWNLOAD] + ": " + ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if settings.MATTERMOST_NOTIFY_SUBTITLEDOWNLOAD:
            self._notify_mattermost(common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD] + " " + ep_name + ": " + lang)

    def notify_update(self, new_version="??"):
        if settings.USE_MATTERMOST:
            update_text = common.notifyStrings[common.NOTIFY_UPDATE_TEXT]
            title = common.notifyStrings[common.NOTIFY_UPDATE]
            self._notify_mattermost(title + " - " + update_text + new_version)

    def notify_login(self, ipaddress=""):
        if settings.USE_MATTERMOST:
            update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
            title = common.notifyStrings[common.NOTIFY_LOGIN]
            self._notify_mattermost(title + " - " + update_text.format(ipaddress))

    def test_notify(self):
        return self._notify_mattermost("This is a test notification from SickChill", force=True)

    def _send_mattermost(self, message=None):
        mattermost_webhook = self.MATTERMOST_WEBHOOK_URL + settings.MATTERMOST_WEBHOOK.replace(self.MATTERMOST_WEBHOOK_URL, "")
        mattermost_icon_emoji = settings.MATTERMOST_ICON_EMOJI
        mattermost_user = settings.MATTERMOST_USERNAME

        logger.info("Sending mattermost webhook message: " + message)
        logger.info("Sending mattermost webhook message to url: " + mattermost_webhook)

        headers = CaseInsensitiveDict({"Content-Type": "application/json"})

        try:
            r = requests.post(
                mattermost_webhook,
                data=json.dumps(dict(text=message, username=mattermost_user, icon_emoji=mattermost_icon_emoji, icon_url=self.MATTERMOST_ICON_URL)),
                headers=headers,
            )
            r.raise_for_status()
        except Exception as error:
            logger.exception(f"Error Sending Mattermost webhook message: {error}")
            return False

        return True

    def _notify_mattermost(self, message="", force=False):
        if not settings.USE_MATTERMOST and not force:
            logger.debug("Notification for Mattermost webhook not enabled, skipping this notification")
            return False

        return self._send_mattermost(message)



import json
import ntpath
import os
import posixpath

from sickchill import logger, settings
from sickchill.helper import episode_num, try_int
from sickchill.helper.exceptions import CantRefreshShowException, CantUpdateShowException
from sickchill.oldbeard import db, subtitles as subtitle_module, ui
from sickchill.oldbeard.common import Overview, Quality, SNATCHED
from sickchill.show.Show import Show
from sickchill.views.common import PageTemplate
from sickchill.views.home import Home, WebRoot
from sickchill.views.routes import Route


@Route("/manage(/?.*)", name="manage:main")
class Manage(Home, WebRoot):
    def index(self):
        t = PageTemplate(rh=self, filename="manage.mako")
        return t.render(
            title=_("Mass Update"),
            header=_("Mass Update"),
            topmenu="manage",
            controller="manage",
            action="index",
        )

    @staticmethod
    def showEpisodeStatuses(indexer_id, whichStatus):
        status_list = [int(whichStatus)]
        if status_list[0] == SNATCHED:
            status_list = Quality.SNATCHED + Quality.SNATCHED_PROPER + Quality.SNATCHED_BEST

        main_db_con = db.DBConnection()
        cur_show_results = main_db_con.select(
            "SELECT season, episode, name FROM tv_episodes WHERE showid = ? AND season != 0 AND status IN ({0})".format(",".join(["?"] * len(status_list))),
            [int(indexer_id)] + status_list,
        )

        result = {}
        for cur_result in cur_show_results:
            cur_season = int(cur_result["season"])
            cur_episode = int(cur_result["episode"])

            if cur_season not in result:
                result[cur_season] = {}

            result[cur_season][cur_episode] = cur_result["name"]

        return json.dumps(result)

    def episodeStatuses(self):
        which_status = self.get_query_argument("whichStatus", None)
        if which_status:
            status_list = [int(which_status)]
            if status_list[0] == SNATCHED:
                status_list = Quality.SNATCHED + Quality.SNATCHED_PROPER + Quality.SNATCHED_BEST
        else:
            status_list = []

        t = PageTemplate(rh=self, filename="manage_episodeStatuses.mako")

        # if we have no status then this is as far as we need to go
        if not status_list:
            return t.render(
                title=_("Episode Overview"),
                header=_("Episode Overview"),
                topmenu="manage",
                show_names=None,
                whichStatus=which_status,
                ep_counts=None,
                sorted_show_ids=None,
                controller="manage",
                action="episodeStatuses",
            )

        main_db_con = db.DBConnection()
        status_results = main_db_con.select(
            "SELECT show_name, tv_shows.indexer_id AS indexer_id FROM tv_episodes, tv_shows WHERE tv_episodes.status IN ({0}) AND season != 0 AND "
            "tv_episodes.showid = tv_shows.indexer_id ORDER BY show_name".format(",".join(["?"] * len(status_list))),
            status_list,
        )

        ep_counts = {}
        show_names = {}
        sorted_show_ids = []
        for cur_status_result in status_results:
            cur_indexer_id = int(cur_status_result["indexer_id"])
            if cur_indexer_id not in ep_counts:
                ep_counts[cur_indexer_id] = 1
            else:
                ep_counts[cur_indexer_id] += 1

            show_names[cur_indexer_id] = cur_status_result["show_name"]
            if cur_indexer_id not in sorted_show_ids:
                sorted_show_ids.append(cur_indexer_id)

        return t.render(
            title=_("Episode Overview"),
            header=_("Episode Overview"),
            topmenu="manage",
            whichStatus=which_status,
            show_names=show_names,
            ep_counts=ep_counts,
            sorted_show_ids=sorted_show_ids,
            controller="manage",
            action="episodeStatuses",
        )

    # noinspection PyUnusedLocal
    def changeEpisodeStatuses(self, oldStatus, newStatus, *args, **kwargs):
        status_list = [int(oldStatus)]
        if status_list[0] == SNATCHED:
            status_list = Quality.SNATCHED + Quality.SNATCHED_PROPER + Quality.SNATCHED_BEST

        to_change = {}
        self.to_change_show = []
        self.to_change_eps = []

        # make a list of all shows and their associated args
        for arg in kwargs:
            indexer_id, what = arg.split("-")

            # we don't care about unchecked checkboxes
            if kwargs[arg] != "on":
                continue

            if indexer_id not in to_change:
                to_change[indexer_id] = []

            to_change[indexer_id].append(what)

        main_db_con = db.DBConnection()
        for cur_indexer_id in to_change:
            self.to_change_show = cur_indexer_id
            # get a list of all the eps we want to change if they just said "all"
            if "all" in to_change[cur_indexer_id]:
                all_eps_results = main_db_con.select(
                    "SELECT season, episode FROM tv_episodes WHERE status IN ({0}) AND season != 0 AND showid = ?".format(",".join(["?"] * len(status_list))),
                    status_list + [cur_indexer_id],
                )
                all_eps = [str(x["season"]) + "x" + str(x["episode"]) for x in all_eps_results]
                self.to_change_eps = all_eps
            else:
                self.to_change_eps = to_change[cur_indexer_id]

            self.setStatus(direct=True)

        return self.redirect("/manage/episodeStatuses/")

    @staticmethod
    def showSubtitleMissed(indexer_id, whichSubs):
        main_db_con = db.DBConnection()
        cur_show_results = main_db_con.select(
            "SELECT season, episode, name, subtitles FROM tv_episodes WHERE showid = ? {0} AND (status LIKE '%4' OR status LIKE '%6') and "
            "location != ''".format(("AND season != 0 ", "")[settings.SUBTITLES_INCLUDE_SPECIALS]),
            [int(indexer_id)],
        )
        result = {}
        for cur_result in cur_show_results:
            if whichSubs == "all":
                if not frozenset(subtitle_module.wanted_languages()).difference(cur_result["subtitles"].split(",")):
                    continue
            elif whichSubs in cur_result["subtitles"]:
                continue

            cur_season = int(cur_result["season"])
            cur_episode = int(cur_result["episode"])

            if cur_season not in result:
                result[cur_season] = {}

            if cur_episode not in result[cur_season]:
                result[cur_season][cur_episode] = {}

            result[cur_season][cur_episode]["name"] = cur_result["name"]

            result[cur_season][cur_episode]["subtitles"] = cur_result["subtitles"]

        return json.dumps(result)

    def subtitleMissed(self):
        which_subs = self.get_body_argument("whichSubs", None)
        t = PageTemplate(rh=self, filename="manage_subtitleMissed.mako")

        if not which_subs:
            return t.render(
                whichSubs=which_subs,
                title=_("Episode Overview"),
                header=_("Episode Overview"),
                topmenu="manage",
                show_names=None,
                ep_counts=None,
                sorted_show_ids=None,
                controller="manage",
                action="subtitleMissed",
            )

        main_db_con = db.DBConnection()
        status_results = main_db_con.select(
            "SELECT show_name, tv_shows.indexer_id as indexer_id, tv_episodes.subtitles subtitles "
            + "FROM tv_episodes, tv_shows "
            + "WHERE tv_shows.subtitles = 1 AND (tv_episodes.status LIKE '%4' OR tv_episodes.status LIKE '%6') AND tv_episodes.season != 0 "
            + "AND tv_episodes.location != '' AND tv_episodes.showid = tv_shows.indexer_id ORDER BY show_name"
        )

        ep_counts = {}
        show_names = {}
        sorted_show_ids = []
        for cur_status_result in status_results:
            if which_subs == "all":
                if not frozenset(subtitle_module.wanted_languages()).difference(cur_status_result["subtitles"].split(",")):
                    continue
            elif which_subs in cur_status_result["subtitles"]:
                continue

            cur_indexer_id = int(cur_status_result["indexer_id"])
            if cur_indexer_id not in ep_counts:
                ep_counts[cur_indexer_id] = 1
            else:
                ep_counts[cur_indexer_id] += 1

            show_names[cur_indexer_id] = cur_status_result["show_name"]
            if cur_indexer_id not in sorted_show_ids:
                sorted_show_ids.append(cur_indexer_id)

        return t.render(
            whichSubs=which_subs,
            show_names=show_names,
            ep_counts=ep_counts,
            sorted_show_ids=sorted_show_ids,
            title=_("Missing Subtitles"),
            header=_("Missing Subtitles"),
            topmenu="manage",
            controller="manage",
            action="subtitleMissed",
        )

    # noinspection PyUnusedLocal
    def downloadSubtitleMissed(self, *args, **kwargs):
        to_download = {}

        # make a list of all shows and their associated args
        for arg in kwargs:
            indexer_id, what = arg.split("-")

            # we don't care about unchecked checkboxes
            if kwargs[arg] != "on":
                continue

            if indexer_id not in to_download:
                to_download[indexer_id] = []

            to_download[indexer_id].append(what)

        for cur_indexer_id in to_download:
            # get a list of all the eps we want to download subtitles if they just said "all"
            if "all" in to_download[cur_indexer_id]:
                main_db_con = db.DBConnection()
                all_eps_results = main_db_con.select(
                    "SELECT season, episode FROM tv_episodes WHERE (status LIKE '%4' OR status LIKE '%6') {0} AND showid = ? AND location != ''".format(
                        ("AND season != 0 ", "")[settings.SUBTITLES_INCLUDE_SPECIALS]
                    ),
                    [cur_indexer_id],
                )
                to_download[cur_indexer_id] = [str(x["season"]) + "x" + str(x["episode"]) for x in all_eps_results]

            for epResult in to_download[cur_indexer_id]:
                season, episode = epResult.split("x")

                show = Show.find(settings.show_list, int(cur_indexer_id))
                show.get_episode(season, episode).download_subtitles()

        return self.redirect("/manage/subtitleMissed/")

    def backlogShow(self):
        show = self.get_query_argument("indexer_id")
        show_object = Show.find(settings.show_list, show)

        if show_object:
            settings.backlogSearchScheduler.action.searchBacklog([show_object])

        return self.redirect("/manage/backlogOverview/")

    def backlogOverview(self):
        t = PageTemplate(rh=self, filename="manage_backlogOverview.mako")

        show_counts = {}
        show_categories = {}
        show_sql_results = {}

        main_db_con = db.DBConnection()
        for current_show in settings.show_list:
            episode_counts = {
                Overview.SKIPPED: 0,
                Overview.WANTED: 0,
                Overview.QUAL: 0,
                Overview.GOOD: 0,
                Overview.UNAIRED: 0,
                Overview.SNATCHED: 0,
                Overview.SNATCHED_PROPER: 0,
                Overview.SNATCHED_BEST: 0,
            }
            episode_categories = {}

            sql_results = main_db_con.select(
                "SELECT status, season, episode, name, airdate FROM tv_episodes WHERE tv_episodes.season IS NOT NULL "
                "AND tv_episodes.showid IN (SELECT tv_shows.indexer_id FROM tv_shows WHERE tv_shows.indexer_id = ? "
                "AND paused = 0) ORDER BY tv_episodes.season DESC, tv_episodes.episode DESC",
                [current_show.indexerid],
            )

            for result in sql_results:
                curEpCat = current_show.get_overview(result["status"], backlog=settings.BACKLOG_MISSING_ONLY)
                if curEpCat:
                    episode_categories["{ep}".format(ep=episode_num(result["season"], result["episode"]))] = curEpCat
                    episode_counts[curEpCat] += 1

            show_counts[current_show.indexerid] = episode_counts
            show_categories[current_show.indexerid] = episode_categories
            show_sql_results[current_show.indexerid] = sql_results

        def show_qual_snatched(show):
            return Quality.splitQuality(show.quality)[1]

        total_wanted = total_qual = total_qual_snatched = 0
        backlog_shows = sorted(
            [
                x
                for x in settings.show_list
                if (
                    show_counts[x.indexerid][Overview.QUAL]
                    or show_counts[x.indexerid][Overview.WANTED]
                    or (0, show_counts[x.indexerid][Overview.SNATCHED])[len(show_qual_snatched(x)) > 0]
                )
            ],
            key=lambda x: x.sort_name,
        )
        for current_show in backlog_shows:
            total_wanted += show_counts[current_show.indexerid][Overview.WANTED]
            total_qual += show_counts[current_show.indexerid][Overview.QUAL]
            if show_qual_snatched(current_show):
                total_qual_snatched += show_counts[current_show.indexerid][Overview.SNATCHED]

        return t.render(
            show_counts=show_counts,
            show_categories=show_categories,
            total_qual=total_qual,
            show_qual_snatched=show_qual_snatched,
            total_wanted=total_wanted,
            total_qual_snatched=total_qual_snatched,
            backlog_shows=backlog_shows,
            show_sql_results=show_sql_results,
            controller="manage",
            action="backlogOverview",
            title=_("Backlog Overview"),
            header=_("Backlog Overview"),
            topmenu="manage",
        )

    @staticmethod
    def __gooey_path(name, method):
        result = getattr(os.path, method)(name)
        if result == name or not result:
            result = getattr(ntpath, method)(name)
            if result == name or not result:
                result = getattr(posixpath, method)(name)

        return result

    # noinspection PyProtectedMember
    def massEdit(self):
        t = PageTemplate(rh=self, filename="manage_massEdit.mako")

        edit = self.get_body_arguments("edit")

        show_list = []
        show_names = []
        for show_id in edit:
            show_object = Show.find(settings.show_list, show_id)
            if show_object:
                show_list.append(show_object)
                show_names.append(show_object.name)

        season_folders_all_same = True
        last_season_folders = None

        paused_all_same = True
        last_paused = None

        default_ep_status_all_same = True
        last_default_ep_status = None

        anime_all_same = True
        last_anime = None

        sports_all_same = True
        last_sports = None

        quality_all_same = True
        last_quality = None

        subtitles_all_same = True
        last_subtitles = None

        scene_all_same = True
        last_scene = None

        air_by_date_all_same = True
        last_air_by_date = None

        root_dir_list = []

        for current_show in show_list:
            show_root_dir = self.__gooey_path(current_show.get_location, "dirname")
            if show_root_dir and show_root_dir != current_show.get_location and show_root_dir not in root_dir_list:
                root_dir_list.append(show_root_dir)

            # if we know they're not all the same then no point even bothering
            if paused_all_same:
                # if we had a value already and this value is different, then they're not all the same
                if last_paused not in (None, current_show.paused):
                    paused_all_same = False
                else:
                    last_paused = current_show.paused

            if default_ep_status_all_same:
                if last_default_ep_status not in (None, current_show.default_ep_status):
                    default_ep_status_all_same = False
                else:
                    last_default_ep_status = current_show.default_ep_status

            if anime_all_same:
                # if we had a value already and this value is different, then they're not all the same
                if last_anime not in (None, current_show.is_anime):
                    anime_all_same = False
                else:
                    last_anime = current_show.anime

            if season_folders_all_same:
                if last_season_folders not in (None, current_show.season_folders):
                    season_folders_all_same = False
                else:
                    last_season_folders = current_show.season_folders

            if quality_all_same:
                if last_quality not in (None, current_show.quality):
                    quality_all_same = False
                else:
                    last_quality = current_show.quality

            if subtitles_all_same:
                if last_subtitles not in (None, current_show.subtitles):
                    subtitles_all_same = False
                else:
                    last_subtitles = current_show.subtitles

            if scene_all_same:
                if last_scene not in (None, current_show.scene):
                    scene_all_same = False
                else:
                    last_scene = current_show.scene

            if sports_all_same:
                if last_sports not in (None, current_show.sports):
                    sports_all_same = False
                else:
                    last_sports = current_show.sports

            if air_by_date_all_same:
                if last_air_by_date not in (None, current_show.air_by_date):
                    air_by_date_all_same = False
                else:
                    last_air_by_date = current_show.air_by_date

        default_ep_status_value = last_default_ep_status if default_ep_status_all_same else None
        paused_value = last_paused if paused_all_same else None
        anime_value = last_anime if anime_all_same else None
        season_folders_value = last_season_folders if season_folders_all_same else None
        quality_value = last_quality if quality_all_same else None
        subtitles_value = last_subtitles if subtitles_all_same else None
        scene_value = last_scene if scene_all_same else None
        sports_value = last_sports if sports_all_same else None
        air_by_date_value = last_air_by_date if air_by_date_all_same else None
        ignore_words_value = None
        prefer_words_value = None
        require_words_value = None

        return t.render(
            edit_list=edit,
            show_names=show_names,
            default_ep_status_value=default_ep_status_value,
            paused_value=paused_value,
            anime_value=anime_value,
            season_folders_value=season_folders_value,
            quality_value=quality_value,
            subtitles_value=subtitles_value,
            scene_value=scene_value,
            sports_value=sports_value,
            air_by_date_value=air_by_date_value,
            root_dir_list=root_dir_list,
            title=_("Mass Edit"),
            header=_("Mass Edit"),
            controller="manage",
            action="massEdit",
            topmenu="manage",
            ignore_words_value=ignore_words_value,
            prefer_words_value=prefer_words_value,
            require_words_value=require_words_value,
        )

    # noinspection PyProtectedMember, PyUnusedLocal
    def massEditSubmit(self):
        paused = self.get_body_argument("paused", None)
        default_ep_status = self.get_body_argument("default_ep_status", None)
        anime = self.get_body_argument("anime", None)
        sports = self.get_body_argument("sports", None)
        scene = self.get_body_argument("scene", None)
        season_folders = self.get_body_argument("season_folders", None)
        quality_preset = self.get_body_argument("quality_preset", None)
        subtitles = self.get_body_argument("subtitles", None)
        air_by_date = self.get_body_argument("air_by_date", None)
        any_qualities = self.get_body_arguments("anyQualities")
        best_qualities = self.get_body_arguments("bestQualities")
        edit = self.get_body_arguments("edit")
        mass_ignore_words = self.get_body_argument("mass_ignore_words", None)
        mass_prefer_words = self.get_body_argument("mass_prefer_words", None)
        mass_require_words = self.get_body_argument("mass_require_words", None)
        ignore_words = self.get_body_argument("ignore_words", None)
        prefer_words = self.get_body_argument("prefer_words", None)
        require_words = self.get_body_argument("require_words", None)

        root_dir_map = {}
        for root_dir in [x for x in self.request.body_arguments if x.startswith("orig_root_dir_")]:
            old_root = self.get_body_argument(root_dir, None)
            new_root = self.get_body_argument(root_dir.replace("orig_root_dir_", "new_root_dir_"), None)
            if old_root is not None and new_root is not None:
                root_dir_map[old_root] = new_root

        errors = []
        for current_show in edit:
            current_show_errors = []
            show_object = Show.find(settings.show_list, current_show)
            if not show_object:
                continue

            show_root_dir = self.__gooey_path(show_object.get_location, "dirname")
            cur_show_dir = self.__gooey_path(show_object.get_location, "basename")
            if show_root_dir and root_dir_map.get(show_root_dir) and show_root_dir != root_dir_map.get(show_root_dir):
                new_show_dir = os.path.join(root_dir_map[show_root_dir], cur_show_dir)
                logger.info(f"For show {show_object.name} changing dir from {show_object.get_location} to {new_show_dir}")
            else:
                new_show_dir = show_object.get_location

            new_paused = ("off", "on")[(paused == "enable", show_object.paused)[paused == "keep"]]
            new_default_ep_status = (default_ep_status, show_object.default_ep_status)[default_ep_status == "keep"]
            new_anime = ("off", "on")[(anime == "enable", show_object.anime)[anime == "keep"]]
            new_sports = ("off", "on")[(sports == "enable", show_object.sports)[sports == "keep"]]
            new_scene = ("off", "on")[(scene == "enable", show_object.scene)[scene == "keep"]]
            new_air_by_date = ("off", "on")[(air_by_date == "enable", show_object.air_by_date)[air_by_date == "keep"]]
            new_season_folders = ("off", "on")[(season_folders == "enable", show_object.season_folders)[season_folders == "keep"]]
            new_subtitles = ("off", "on")[(subtitles == "enable", show_object.subtitles)[subtitles == "keep"]]

            # new mass words update section
            if ignore_words == "new":
                new_ignore_words = mass_ignore_words
            elif ignore_words == "clear":
                new_ignore_words = ""
            else:
                new_ignore_words = show_object.rls_ignore_words

            if require_words == "new":
                new_require_words = mass_require_words
            elif require_words == "clear":
                new_require_words = ""
            else:
                new_require_words = show_object.rls_require_words

            if prefer_words == "new":
                new_prefer_words = mass_prefer_words
            elif prefer_words == "clear":
                new_prefer_words = ""
            else:
                new_prefer_words = show_object.rls_prefer_words

            if quality_preset == "keep":
                any_qualities, best_qualities = Quality.splitQuality(show_object.quality)
            elif try_int(quality_preset, None):
                best_qualities = []

            exceptions_list = []

            current_show_errors += self.editShow(
                current_show,
                new_show_dir,
                any_qualities,
                best_qualities,
                exceptions_list,
                defaultEpStatus=new_default_ep_status,
                season_folders=new_season_folders,
                paused=new_paused,
                sports=new_sports,
                subtitles=new_subtitles,
                rls_ignore_words=new_ignore_words,
                rls_prefer_words=new_prefer_words,
                rls_require_words=new_require_words,
                anime=new_anime,
                scene=new_scene,
                air_by_date=new_air_by_date,
                directCall=True,
            )

            if current_show_errors:
                logger.exception(f"Errors: {current_show_errors}")
                errors.append(f"<b>{show_object.name}:</b>\n<ul>" + " ".join([f"<li>{error}</li>" for error in current_show_errors]) + "</ul>")

        if errors:
            ui.notifications.error(
                _("{num_errors:d} error{plural} while saving changes:").format(num_errors=len(errors), plural="" if len(errors) == 1 else "s"), " ".join(errors)
            )

        return self.redirect("/manage/")

    def massUpdate(self):
        update = self.get_body_arguments("update")
        refresh = self.get_body_arguments("refresh")
        rename = self.get_body_arguments("rename")
        subtitle = self.get_body_arguments("subtitle")
        delete = self.get_body_arguments("delete")
        remove = self.get_body_arguments("remove")
        metadata = self.get_body_arguments("metadata")
        edit = self.get_body_arguments("edit")

        errors = []
        refreshes = []
        updates = []
        renames = []
        subtitles = []

        for curShowID in set(update + refresh + rename + subtitle + delete + remove + metadata):
            if curShowID == "":
                continue

            show_object = Show.find(settings.show_list, curShowID)
            if not show_object:
                continue

            if curShowID in delete:
                Show.delete(show_object, True)
                # don't do anything else if it's being deleted
                continue

            if curShowID in remove:
                Show.delete(show_object)
                # don't do anything else if it's being removed
                continue

            if curShowID in update:
                try:
                    Show.update(show_object, True)
                    updates.append(show_object.name)
                except CantUpdateShowException as error:
                    errors.append(_("Unable to update show: {exception_format}").format(exception_format=error))

            # don't bother refreshing shows that were updated anyway
            if curShowID in refresh and curShowID not in update:
                try:
                    Show.refresh(show_object, force=True)
                    refreshes.append(show_object.name)
                except CantRefreshShowException as error:
                    errors.append(_("Unable to refresh show {show_name}: {exception_format}").format(show_name=show_object.name, exception_format=error))

            if curShowID in rename:
                settings.showQueueScheduler.action.rename_show_episodes(show_object)
                renames.append(show_object.name)

            if curShowID in subtitle:
                settings.showQueueScheduler.action.download_subtitles(show_object)
                subtitles.append(show_object.name)

        if errors:
            ui.notifications.error(_("Errors encountered"), "<br >\n".join(errors))

        message_detail = ""

        if updates:
            message_detail += "<br><b>" + _("Updates") + "</b><br><ul><li>"
            message_detail += "</li><li>".join(updates)
            message_detail += "</li></ul>"

        if refreshes:
            message_detail += "<br><b>" + _("Refreshes") + "</b><br><ul><li>"
            message_detail += "</li><li>".join(refreshes)
            message_detail += "</li></ul>"

        if renames:
            message_detail += "<br><b>" + _("Renames") + "</b><br><ul><li>"
            message_detail += "</li><li>".join(renames)
            message_detail += "</li></ul>"

        if subtitles:
            message_detail += "<br><b>" + _("Subtitles") + "</b><br><ul><li>"
            message_detail += "</li><li>".join(subtitles)
            message_detail += "</li></ul>"

        if updates + refreshes + renames + subtitles:
            ui.notifications.message(_("The following actions were queued") + ":", message_detail)

        if edit:
            return self.massEdit()

        return self.redirect("/manage/")

    def failedDownloads(self):
        remove = self.get_body_arguments("remove[]")
        limit = self.get_argument("limit", "100")
        failed_db_con = db.DBConnection("failed.db")

        if limit == "0":
            sql_results = failed_db_con.select("SELECT * FROM failed")
        else:
            sql_results = failed_db_con.select("SELECT * FROM failed LIMIT ?", [limit])

        for release in remove:
            failed_db_con.action("DELETE FROM failed WHERE failed.release = ?", [release])

        if remove:
            return self.redirect("/manage/failedDownloads/")

        t = PageTemplate(rh=self, filename="manage_failedDownloads.mako")

        return t.render(
            limit=limit,
            failedResults=sql_results,
            title=_("Failed Downloads"),
            header=_("Failed Downloads"),
            topmenu="manage",
            controller="manage",
            action="failedDownloads",
        )



import datetime
import os
import re
import threading
import traceback
from typing import TYPE_CHECKING

import sickchill.oldbeard.name_cache
import sickchill.oldbeard.providers
from sickchill import logger, settings
from sickchill.helper.exceptions import AuthException
from sickchill.providers.GenericProvider import GenericProvider
from sickchill.show.History import History

if TYPE_CHECKING:  # pragma: no cover
    from sickchill.oldbeard.classes import SearchResult

from . import clients, common, db, helpers, notifiers, nzbget, nzbSplitter, sab, show_name_helpers, ui
from .common import MULTI_EP_RESULT, Quality, SEASON_RESULT, SNATCHED, SNATCHED_BEST, SNATCHED_PROPER


def _download_result(result: "SearchResult"):
    """
    Downloads a result to the appropriate black hole folder.

    :param result: SearchResult instance to download.
    :return: boolean, True on success
    """

    result_was_downloaded = False

    result_provider = result.provider
    if result_provider is None:
        logger.exception("Invalid provider name - this is a coding error, report it please")
        return result_was_downloaded

    # nzbs/torrents with a URL can just be downloaded from the provider
    if result.is_torrent or result.is_nzb:
        result_was_downloaded = result_provider.download_result(result)
    # if it's an nzb data result
    elif result.is_nzbdata:
        # get the final file path to the nzb
        filename = os.path.join(settings.NZB_DIR, f"{result.name}.nzb")
        logger.info(f"Saving NZB to {filename}")

        # save the data to disk
        try:
            with open(filename, "w") as fileOut:
                fileOut.write(result.extraInfo[0])

            result_was_downloaded = True
            helpers.chmodAsParent(filename)

        except EnvironmentError as error:
            logger.exception(f"Error trying to save NZB to black hole: {error}")
            result_was_downloaded = False
    else:
        logger.exception("Invalid provider type - this is a coding error, report it please")
        result_was_downloaded = False

    return result_was_downloaded


def snatch_episode(result: "SearchResult", end_status=SNATCHED):
    """
    Contains the internal logic necessary to actually "snatch" a result that
    has been found.

    :param result: SearchResult instance to be snatched.
    :param end_status: the episode status that should be used for the episode object once it's snatched.
    :return: boolean, True on success
    """

    if result is None:
        return False

    if settings.ALLOW_HIGH_PRIORITY:
        # if it aired recently make it high priority
        for episode in result.episodes:
            if datetime.date.today() - episode.airdate <= datetime.timedelta(days=7):
                result.priority = 1

    end_status = SNATCHED_PROPER if re.search(r"\b(proper|repack|real)\b", result.name, re.I) else end_status

    # Torrents can be sent to clients or saved to disk
    if result.is_torrent:
        # torrents are saved to disk when blackhole mode
        if settings.TORRENT_METHOD == "blackhole":
            snatched_result = _download_result(result)
        else:
            if not result.content and not result.url.startswith("magnet"):
                if result.provider.login():
                    result.content = result.provider.get_url(result.url, returns="content")

            if result.content or result.url.startswith("magnet"):
                client = clients.getClientInstance(settings.TORRENT_METHOD)()
                snatched_result = client.sendTORRENT(result)
            else:
                logger.warning("Torrent file content is empty")
                # TODO: This is broken!!
                # History().log_failed(result.episodes, result.name, result.provider)
                History().log_failed(result.episodes[0], result.name, result.provider.name)  # This one seems to work
                snatched_result = False
    # NZBs can be sent straight to SAB or saved to disk
    elif result.is_nzb or result.is_nzbdata:
        if settings.NZB_METHOD == "blackhole":
            snatched_result = _download_result(result)
        elif settings.NZB_METHOD == "sabnzbd":
            snatched_result = sab.send_nzb(result)
        elif settings.NZB_METHOD == "nzbget":
            is_proper = True if end_status == SNATCHED_PROPER else False
            snatched_result = nzbget.send_nzb(result, is_proper)
        elif settings.NZB_METHOD == "download_station":
            client = clients.getClientInstance(settings.NZB_METHOD)(settings.SYNOLOGY_DSM_HOST, settings.SYNOLOGY_DSM_USERNAME, settings.SYNOLOGY_DSM_PASSWORD)
            snatched_result = client.send_nzb(result)
        else:
            logger.exception(f"Unknown NZB action specified in config: {settings.NZB_METHOD}")
            snatched_result = False
    else:
        logger.exception(f"Unknown result type, unable to download it ({result.result_type})")
        snatched_result = False

    if not snatched_result:
        return False

    ui.notifications.message("Episode snatched", result.name)
    History().log_snatch(result)

    # don't notify when we re-download an episode
    sql_l = []
    trakt_data = []
    for current_episode in result.episodes:
        with current_episode.lock:
            if is_first_best_match(result):
                current_episode.status = Quality.compositeStatus(SNATCHED_BEST, result.quality)
            else:
                current_episode.status = Quality.compositeStatus(end_status, result.quality)

            sql_l.append(current_episode.get_sql())

        if current_episode.status not in Quality.DOWNLOADED:
            # noinspection PyBroadException
            try:
                notifiers.notify_snatch(f"{current_episode.naming_pattern('%SN - %Sx%0E - %EN - %QN')} from {result.provider.name}")
            except Exception:
                # Without this, when notification fail, it crashes the snatch thread and SC will
                # keep snatching until notification is sent
                logger.debug("Failed to send snatch notification")

            trakt_data.append((current_episode.season, current_episode.episode))

    data = notifiers.trakt_notifier.trakt_episode_data_generate(trakt_data)

    if settings.USE_TRAKT and settings.TRAKT_SYNC_WATCHLIST:
        logger.debug(f"Add episodes, showid: indexerid {result.show.indexerid}, Title {result.show.name} to Traktv Watchlist")
        if data:
            notifiers.trakt_notifier.update_watchlist(result.show, data_episode=data, update="add")

    if sql_l:
        main_db_con = db.DBConnection()
        main_db_con.mass_action(sql_l)

    return True


def pick_best_result(results, show):
    """
    Find the best result out of a list of search results for a show

    :param results: list of result objects
    :param show: Shows we check for
    :return: best result object
    """
    results = results if isinstance(results, list) else [results]

    logger.debug(f"Picking the best result out of {[x.name for x in results]}")

    picked_result = None

    # order the list so that preferred releases are at the top
    results.sort(key=lambda ep: show_name_helpers.hasPreferredWords(ep.name, ep.show), reverse=True)

    # find the best result for the current episode
    for result in results:
        if show and result.show is not show:
            continue

        # build the black And white list
        if show.is_anime:
            if not show.release_groups.is_valid(result):
                continue

        logger.info(f"Quality of {result.name} is {Quality.qualityStrings[result.quality]}")

        allowed_qualities, preferred_qualities = Quality.splitQuality(show.quality)

        if result.quality not in allowed_qualities + preferred_qualities:
            logger.debug(f"{result.name} is a quality we know we don't want, rejecting it")
            continue

        if not show_name_helpers.filter_bad_releases(result.name, parse=False, show=show):
            continue

        if hasattr(result, "size"):
            if settings.USE_FAILED_DOWNLOADS and History().has_failed(result.name, result.size, result.provider.name):
                logger.info(f"{result.name} has previously failed, rejecting it")
                continue

        if not picked_result:
            picked_result = result
        elif result.quality in preferred_qualities and (picked_result.quality < result.quality or picked_result.quality not in preferred_qualities):
            picked_result = result
        elif result.quality in allowed_qualities and picked_result.quality not in preferred_qualities and picked_result.quality < result.quality:
            picked_result = result
        elif picked_result.quality == result.quality:
            if "proper" in result.name.lower() or "real" in result.name.lower() or "repack" in result.name.lower():
                logger.info(f"Preferring {result.name} (repack/proper/real over nuked)")
                picked_result = result
            elif "internal" in picked_result.name.lower() and "internal" not in result.name.lower():
                logger.info(f"Preferring {result.name} (normal instead of internal)")
                picked_result = result
            elif "xvid" in picked_result.name.lower() and "x264" in result.name.lower():
                logger.info(f"Preferring {result.name} (x264 over xvid)")
                picked_result = result

    if picked_result:
        logger.debug(f"Picked {picked_result.name} as the best")
    else:
        logger.debug("No result picked.")

    return picked_result


def is_final_result(result: "SearchResult"):
    """
    Checks if the given result is good enough quality that we can stop searching for other ones.

    :param result: quality to check
    :return: True if the result is the highest quality in both of the quality lists else False
    """

    logger.debug(f"Checking if we should keep searching after we've found {result.name}")

    show_obj = result.episodes[0].show

    any_qualities, best_qualities = Quality.splitQuality(show_obj.quality)

    # if there is a re-download that's higher than this then we definitely need to keep looking
    if best_qualities and result.quality < max(best_qualities):
        return False

    # if it does not match the shows black and white list its no good
    elif show_obj.is_anime and show_obj.release_groups.is_valid(result):
        return False

    # if there's no re-download that's higher (above) and this is the highest initial download then we're good
    elif any_qualities and result.quality in any_qualities:
        return True

    elif best_qualities and result.quality == max(best_qualities):
        return True

    # if we got here than it's either not on the lists, they're empty, or it's lower than the highest required
    else:
        return False


def is_first_best_match(result: "SearchResult"):
    """
    Checks if the given result is the best match and if we want to stop searching providers here.

    :param result: to check
    :return: True if the result is the best quality match else False
    """

    logger.debug(f"Checking if we should stop searching for a better quality for for episode {result.name}")

    show_obj = result.episodes[0].show

    any_qualities_, best_qualities = Quality.splitQuality(show_obj.quality)

    return result.quality in best_qualities if best_qualities else False


def wanted_episodes(show, from_date):
    """
    Get a list of episodes that we want to download
    :param show: Show these episodes are from
    :param from_date: Search from a certain date
    :return: list of wanted episodes
    """
    wanted = []
    if show.paused:
        logger.debug(f"Not checking for episodes of {show.name} because the show is paused")
        return wanted

    allowed_qualities, preferred_qualities = common.Quality.splitQuality(show.quality)
    all_qualities = list(set(allowed_qualities + preferred_qualities))

    logger.debug(f"Seeing if we need anything from {show.name}")
    con = db.DBConnection()

    sql_results = con.select(
        "SELECT status, season, episode FROM tv_episodes WHERE showid = ? AND season > 0 and airdate > ?", [show.indexerid, from_date.toordinal()]
    )

    # check through the list of statuses to see if we want any
    for result in sql_results:
        status, quality = common.Quality.splitCompositeStatus(int(result["status"] or -1))
        if status not in {common.WANTED, common.DOWNLOADED, common.SNATCHED, common.SNATCHED_PROPER}:
            continue

        if status != common.WANTED:
            if preferred_qualities:
                if quality in preferred_qualities:
                    continue
            elif quality in allowed_qualities:
                continue

        episode_object = show.get_episode(result["season"], result["episode"])
        episode_object.wantedQuality = [i for i in all_qualities if i > quality and i != common.Quality.UNKNOWN]
        wanted.append(episode_object)

    return wanted


def search_for_needed_episodes():
    """
    Check providers for details on wanted episodes

    :return: episodes we have a search hit for
    """
    found_results = {}

    did_search = False

    show_list = settings.show_list
    from_date = datetime.date.min
    episodes = []

    for curShow in show_list:
        if not curShow.paused:
            sickchill.oldbeard.name_cache.build_name_cache(curShow)
            episodes.extend(wanted_episodes(curShow, from_date))

    if not episodes:
        # nothing wanted so early out, ie: avoid whatever arbitrarily
        # complex thing a provider cache update entails, for example,
        # reading rss feeds
        logger.info("No episodes needed.")
        return list(found_results.values())

    # noinspection DuplicatedCode
    original_thread_name = threading.current_thread().name

    providers = [x for x in sickchill.oldbeard.providers.sorted_provider_list(settings.RANDOMIZE_PROVIDERS) if x.is_active and x.enable_daily and x.can_daily]
    for curProvider in providers:
        threading.current_thread().name = f"{original_thread_name} :: [{curProvider.name}]"
        curProvider.cache.update_cache()

    for curProvider in providers:
        threading.current_thread().name = f"{original_thread_name} :: [{curProvider.name}]"
        try:
            found_rss_results = curProvider.search_rss(episodes)
        except AuthException as error:
            logger.warning(f"Authentication error: {error}")
            continue
        except Exception as error:
            logger.exception(f"Error while searching {curProvider.name}, skipping: {error}")
            logger.debug(traceback.format_exc())
            continue

        did_search = True

        # pick a single result for each episode, respecting existing results
        for current_episode in found_rss_results:
            if not current_episode.show or current_episode.show.paused:
                logger.debug(f"Skipping {current_episode.pretty_name} because the show is paused ")
                continue

            best_result = pick_best_result(found_rss_results[current_episode], current_episode.show)

            # if all results were rejected move on to the next episode
            if not best_result:
                logger.debug(f"All found results for {current_episode.pretty_name} were rejected.")
                continue

            # if it's already in the list (from another provider) and the newly found quality is no better, then skip it
            if current_episode in found_results and best_result.quality <= found_results[current_episode].quality:
                continue

            found_results[current_episode] = best_result

    threading.current_thread().name = original_thread_name

    if not did_search:
        logger.info("No NZB/Torrent providers found or enabled in the sickchill config for daily searches. Please check your settings.")

    return list(found_results.values())


# noinspection PyPep8Naming
def search_providers(show, episodes, manual=False, downCurQuality=False):
    """
    Walk providers for information on shows

    :param show: Show we are looking for
    :param episodes: Episodes we hope to find
    :param manual: Boolean, is this a manual search?
    :param downCurQuality: Boolean, should we re-download currently available quality file
    :return: results for search
    """
    found_results = {}
    final_results = []

    did_search = False

    # build name cache for show
    sickchill.oldbeard.name_cache.build_name_cache(show)

    # noinspection DuplicatedCode
    original_thread_name = threading.current_thread().name

    providers = [
        x for x in sickchill.oldbeard.providers.sorted_provider_list(settings.RANDOMIZE_PROVIDERS) if x.is_active and x.can_backlog and x.enable_backlog
    ]
    for curProvider in providers:
        threading.current_thread().name = f"{original_thread_name} :: [{curProvider.name}]"
        curProvider.cache.update_cache()

    threading.current_thread().name = original_thread_name

    for curProvider in providers:
        threading.current_thread().name = f"{original_thread_name} :: [{curProvider.name}]"

        if curProvider.anime_only and not show.is_anime:
            logger.debug(f"{show.name} is not an anime, skipping")
            continue

        found_results[curProvider.name] = {}

        search_count = 0
        search_mode = curProvider.search_mode

        # Always search for episode when manually searching when in season
        if search_mode == "season" and manual is True:
            search_mode = "episode"

        while True:
            search_count += 1

            logger.info(
                _("Performing {episode_or_season} search for {show}").format(
                    episode_or_season=(_("season pack"), _("episode"))[search_mode == "episode"], show=show.name
                )
            )

            try:
                search_results = curProvider.find_search_results(show, episodes, search_mode, manual, downCurQuality)
            except AuthException as error:
                logger.warning(f"Authentication error: {error}")
                break
            except Exception as error:
                logger.exception(f"Exception while searching {curProvider.name}. Error: {error}")
                logger.debug(traceback.format_exc())
                break

            did_search = True

            if search_results:
                # make a list of all the results for this provider
                for current_episode in search_results:
                    if current_episode in found_results[curProvider.name]:
                        found_results[curProvider.name][current_episode] += search_results[current_episode]
                    else:
                        found_results[curProvider.name][current_episode] = search_results[current_episode]

                break
            elif search_count == 2 or not curProvider.search_fallback:
                break

            if search_mode == "season":
                logger.debug("Fallback episode search initiated")
                search_mode = "episode"
            else:
                logger.debug("Fallback season pack search initiate")
                search_mode = "season"

        # skip to next provider if we have no results to process
        if not found_results[curProvider.name]:
            continue

        # pick the best season NZB
        best_season_result = None
        if SEASON_RESULT in found_results[curProvider.name]:
            best_season_result = pick_best_result(found_results[curProvider.name][SEASON_RESULT], show)

        highest_quality_overall = 0
        for cur_episode in found_results[curProvider.name]:
            for cur_result in found_results[curProvider.name][cur_episode]:
                if cur_result.quality != Quality.UNKNOWN and cur_result.quality > highest_quality_overall:
                    highest_quality_overall = cur_result.quality
        logger.debug(f"The highest quality of any match is {Quality.qualityStrings[highest_quality_overall]}")

        # see if every episode is wanted
        if best_season_result:
            searched_seasons = {str(x.season) for x in episodes}

            # get the quality of the season nzb
            season_quality = best_season_result.quality
            logger.info(f"The quality of the season {best_season_result.provider.provider_type} is {Quality.qualityStrings[season_quality]}")

            main_db_con = db.DBConnection()
            all_episodes = [
                int(x["episode"])
                for x in main_db_con.select(
                    f"SELECT episode FROM tv_episodes WHERE showid = ? AND ( season IN ( {','.join(['?'] * len(searched_seasons))} ) )",
                    [show.indexerid] + list(searched_seasons),
                )
            ]

            logger.info(f"Executed query: [SELECT episode FROM tv_episodes WHERE showid = {show.indexerid} AND season in {searched_seasons}]")
            logger.debug(f"Episode list: {all_episodes}")

            all_wanted = True
            some_wanted = False
            for episode_number in all_episodes:
                for season in (x.season for x in episodes):
                    if not show.want_episode(season, episode_number, season_quality, downCurQuality):
                        all_wanted = False
                    else:
                        some_wanted = True

            # if we need every ep in the season and there's nothing better, then just download this and be done with it (unless single episodes are preferred)
            if all_wanted and best_season_result.quality == highest_quality_overall:
                logger.info(f"Every ep in this season is needed, downloading the whole {best_season_result.provider.provider_type} {best_season_result.name}")
                episode_objects = []
                for episode_number in all_episodes:
                    for season in {x.season for x in episodes}:
                        episode_objects.append(show.get_episode(season, episode_number))
                best_season_result.episodes = episode_objects

                # Remove provider from thread name before return results
                threading.current_thread().name = original_thread_name

                return [best_season_result]

            elif not some_wanted:
                logger.info(f"No eps from this season are wanted at this quality, ignoring the result of {best_season_result.name}")

            else:
                if best_season_result.result_type != GenericProvider.TORRENT:
                    logger.debug("Breaking apart the NZB and adding the individual ones to our results")

                    # if not, break it apart and add them as the lowest priority results
                    split_results = nzbSplitter.split_result(best_season_result)
                    episode_number = -1
                    for current_result in split_results:
                        if len(current_result.episodes) == 1:
                            episode_number = current_result.episodes[0].episode
                        elif len(current_result.episodes) > 1:
                            episode_number = MULTI_EP_RESULT

                        if episode_number >= 0 and episode_number in found_results[curProvider.name]:
                            found_results[curProvider.name][episode_number].append(current_result)
                        else:
                            found_results[curProvider.name][episode_number] = [current_result]

                # If this is a torrent all we can do is leech the entire torrent, user will have to select which eps not do download in his torrent client
                else:
                    # Season result from Torrent Provider must be a full-season torrent, creating multi-ep result for it.
                    logger.info(
                        "Adding multi-ep result for full-season torrent. Set the episodes you don't want to 'don't download' in your torrent client if desired!"
                    )
                    episode_objects = []
                    for episode_number in all_episodes:
                        for season in {x.season for x in episodes}:
                            episode_objects.append(show.get_episode(season, episode_number))
                    best_season_result.episodes = episode_objects

                    if MULTI_EP_RESULT in found_results[curProvider.name]:
                        found_results[curProvider.name][MULTI_EP_RESULT].append(best_season_result)
                    else:
                        found_results[curProvider.name][MULTI_EP_RESULT] = [best_season_result]

        # go through multi-ep results and see if we really want them or not, get rid of the rest
        multi_results = {}
        if MULTI_EP_RESULT in found_results[curProvider.name]:
            for _multiResult in found_results[curProvider.name][MULTI_EP_RESULT]:
                logger.debug(f"Seeing if we want to bother with multi-episode result {_multiResult.name}")

                # Filter result by ignore/required/whitelist/blacklist/quality, etc
                multi_result = pick_best_result(_multiResult, show)
                if not multi_result:
                    continue

                # see how many of the eps that this result covers aren't covered by single results
                needed_episodes = []
                unneeded_episodes = []
                for epObj in multi_result.episodes:
                    # if we have results for the episode
                    if epObj.episode in found_results[curProvider.name] and len(found_results[curProvider.name][epObj.episode]) > 0:
                        unneeded_episodes.append(epObj.episode)
                    else:
                        needed_episodes.append(epObj.episode)

                logger.debug(f"Single-ep check result is needed_episodes: {needed_episodes}, unneeded_episodes: {unneeded_episodes}")

                if not needed_episodes:
                    logger.debug("All of these episodes were covered by single episode results, ignoring this multi-episode result")
                    continue

                # check if these eps are already covered by another multi-result
                needed_multiepisodes = []
                unneeded_multiepisodes = []
                for epObj in multi_result.episodes:
                    if epObj.episode in multi_results:
                        unneeded_multiepisodes.append(epObj.episode)
                    else:
                        needed_multiepisodes.append(epObj.episode)

                logger.debug(f"Multi-ep check result is needed_multiepisodes: {needed_multiepisodes}, unneeded_multiepisodes: {unneeded_multiepisodes}")

                if not needed_multiepisodes:
                    logger.debug("All of these episodes were covered by another multi-episode nzbs, ignoring this multi-ep result")
                    continue

                # don't bother with the single result if we're going to get it with a multi result
                for epObj in multi_result.episodes:
                    multi_results[epObj.episode] = multi_result
                    if epObj.episode in found_results[curProvider.name]:
                        logger.debug(
                            "A needed multi-episode result overlaps with a single-episode result for ep #"
                            + f"{epObj.episode}"
                            + ", removing the single-episode results from the list"
                        )
                        del found_results[curProvider.name][epObj.episode]

        # of all the single ep results narrow it down to the best one for each episode
        final_results += set(multi_results.values())
        for current_episode in found_results[curProvider.name]:
            if current_episode in (MULTI_EP_RESULT, SEASON_RESULT):
                continue

            if not found_results[curProvider.name][current_episode]:
                continue

            # if all results were rejected move on to the next episode
            best_result = pick_best_result(found_results[curProvider.name][current_episode], show)
            if not best_result:
                continue

            # add result if it's not a duplicate and
            found = False
            for i, result in enumerate(final_results):
                for best_result_episode in best_result.episodes:
                    if best_result_episode in result.episodes:
                        if result.quality < best_result.quality:
                            final_results.pop(i)
                        else:
                            found = True
            if not found:
                final_results += [best_result]

        # check that we got all the episodes we wanted first before doing a match and snatch
        wanted_episode_count = 0
        for wanted_episode in episodes:
            for result in final_results:
                if wanted_episode in result.episodes and is_final_result(result):
                    wanted_episode_count += 1

        # make sure we search every provider for results unless we found everything we wanted
        if wanted_episode_count == len(episodes):
            break

    if not did_search:
        logger.info("No NZB/Torrent providers found or enabled in the sickchill config for backlog searches. Please check your settings.")

    # Remove provider from thread name before return results
    threading.current_thread().name = original_thread_name
    return final_results



import os
import traceback
from collections import namedtuple
from operator import attrgetter
from typing import overload, Tuple, Union

import dateutil.parser

import sickchill
from sickchill import logger, settings
from sickchill.helper.common import sanitize_filename
from sickchill.helper.exceptions import (
    CantRefreshShowException,
    CantRemoveShowException,
    CantUpdateShowException,
    EpisodeDeletedException,
    MultipleShowObjectsException,
    ShowDirectoryNotFoundException,
)
from sickchill.oldbeard.trakt_api import TraktAPI
from sickchill.show.Show import Show
from sickchill.tv import TVShow

from . import generic_queue, name_cache, notifiers, scene_numbering, ui
from .blackandwhitelist import BlackAndWhiteList
from .common import WANTED
from .helpers import chmodAsParent, makeDir, sortable_name


class ShowQueue(generic_queue.GenericQueue):
    def __init__(self):
        super(ShowQueue, self).__init__()
        # noinspection SpellCheckingInspection
        self.queue_name = "SHOWQUEUE"

    def _is_in_queue(self, show: "TVShow", actions) -> bool:
        if not show:
            return False

        return show.indexerid in (x.show.indexerid if x.show else 0 for x in self.queue if x.action_id in actions)

    def _actions_in_queue(self, show: "TVShow", actions) -> bool:
        return self.currentItem is not None and show == self.currentItem.show and self.currentItem.action_id in actions

    # def is_in_add_queue(self, show: "TVShow") -> bool:
    #     return self._isInQueue(show, (ShowQueueActions.ADD,))

    def is_in_update_queue(self, show: "TVShow") -> bool:
        return self._is_in_queue(show, (ShowQueueActions.UPDATE, ShowQueueActions.FORCEUPDATE))

    def is_in_refresh_queue(self, show: "TVShow") -> bool:
        return self._is_in_queue(show, (ShowQueueActions.REFRESH,))

    def is_in_rename_queue(self, show: "TVShow") -> bool:
        return self._is_in_queue(show, (ShowQueueActions.RENAME,))

    def is_in_remove_queue(self, show: "TVShow") -> bool:
        return self._is_in_queue(show, (ShowQueueActions.REMOVE,))

    def is_in_subtitle_queue(self, show: "TVShow") -> bool:
        return self._is_in_queue(show, (ShowQueueActions.SUBTITLE,))

    def is_being_added(self, show: "TVShow") -> bool:
        return self._actions_in_queue(show, (ShowQueueActions.ADD,))

    def is_being_updated(self, show: "TVShow") -> bool:
        return self._actions_in_queue(show, (ShowQueueActions.UPDATE, ShowQueueActions.FORCEUPDATE))

    def is_being_refreshed(self, show: "TVShow") -> bool:
        return self._actions_in_queue(show, (ShowQueueActions.REFRESH,))

    def is_being_renamed(self, show: "TVShow") -> bool:
        return self._actions_in_queue(show, (ShowQueueActions.RENAME,))

    def is_being_removed(self, show: "TVShow") -> bool:
        return self._actions_in_queue(show, (ShowQueueActions.REMOVE,))

    def is_being_subtitled(self, show: "TVShow") -> bool:
        return self._actions_in_queue(show, (ShowQueueActions.SUBTITLE,))

    @property
    def loading_show_list(self) -> {"ShowQueueItem"}:
        return {x for x in self.queue + [self.currentItem] if x and x.is_loading}

    def update_show(self, show: "TVShow", force: bool = False) -> "ShowQueueItem":
        if self.is_being_added(show):
            raise CantUpdateShowException(f"{show.name} is still being added, wait until it is finished before you update.")

        if self.is_being_updated(show):
            raise CantUpdateShowException(f"{show.name} is already being updated by Post-processor or manually started, can't update again until it's done.")

        if self.is_in_update_queue(show):
            raise CantUpdateShowException(
                f"{show.name} is in process of being updated by Post-processor or manually started, can't update again until it's done."
            )

        queue_item_obj = QueueItemUpdate(show, force=force)
        self.add_item(queue_item_obj)
        return queue_item_obj

    def refresh_show(self, show: "TVShow", force=False) -> Union["ShowQueueItem", None]:
        if self.is_being_refreshed(show) and not force:
            raise CantRefreshShowException("This show is already being refreshed, not refreshing again.")

        if self.is_being_updated(show) or self.is_in_update_queue(show):
            logger.debug(
                "A refresh was attempted but there is already an update queued or in progress. Updates do a refresh at the end so I'm skipping this request."
            )
            return

        if show.paused and not force:
            logger.debug(f"Skipping show [{show.name}] because it is paused.")
            return

        logger.debug(f"Queueing show refresh for {show.name}")

        queue_item_obj = QueueItemRefresh(show, force=force)
        self.add_item(queue_item_obj)
        return queue_item_obj

    def rename_show_episodes(self, show: "TVShow") -> "ShowQueueItem":
        queue_item_obj = QueueItemRename(show)
        self.add_item(queue_item_obj)
        return queue_item_obj

    def download_subtitles(self, show: "TVShow") -> "ShowQueueItem":
        queue_item_obj = QueueItemSubtitle(show)
        self.add_item(queue_item_obj)
        return queue_item_obj

    def add_show(
        self,
        indexer,
        indexer_id,
        show_dir,
        default_status=None,
        quality=None,
        season_folders=None,
        lang=None,
        subtitles=None,
        subtitles_sc_metadata=None,
        anime=None,
        scene=None,
        paused=None,
        blacklist=None,
        whitelist=None,
        default_status_after=None,
        root_dir=None,
    ) -> "ShowQueueItem":
        if lang is None:
            lang = settings.INDEXER_DEFAULT_LANGUAGE

        if default_status_after is None:
            default_status_after = settings.STATUS_DEFAULT_AFTER

        queue_item_obj = QueueItemAdd(
            indexer,
            indexer_id,
            show_dir,
            default_status,
            quality,
            season_folders,
            lang,
            subtitles,
            subtitles_sc_metadata,
            anime,
            scene,
            paused,
            blacklist,
            whitelist,
            default_status_after,
            root_dir,
        )

        self.add_item(queue_item_obj)
        return queue_item_obj

    def remove_show(self, show: "TVShow", full: bool = False) -> "ShowQueueItem":
        if not show:
            raise CantRemoveShowException("Failed removing show: Show does not exist")

        if not hasattr(show, "indexerid"):
            raise CantRemoveShowException("Failed removing show: Show does not have an indexer id")

        if self._is_in_queue(show, (ShowQueueActions.REMOVE,)):
            raise CantRemoveShowException(f"{show.name} is already queued to be removed")

        # remove other queued actions for this show.
        for item in self.queue:
            if item and item.show and item != self.currentItem and show.indexerid == item.show.indexerid:
                self.queue.remove(item)

        queue_item_obj = QueueItemRemove(show=show, full=full)
        self.add_item(queue_item_obj)
        return queue_item_obj


class ShowQueueActions(object):
    def __init__(self):
        pass

    REFRESH = 1
    ADD = 2
    UPDATE = 3
    # noinspection SpellCheckingInspection
    FORCEUPDATE = 4
    RENAME = 5
    SUBTITLE = 6
    REMOVE = 7

    names = {
        REFRESH: _("Refresh"),
        ADD: _("Add"),
        UPDATE: _("Update"),
        FORCEUPDATE: _("Force Update"),
        RENAME: _("Rename"),
        SUBTITLE: _("Subtitle"),
        REMOVE: _("Remove Show"),
    }


class ShowQueueItem(generic_queue.QueueItem):
    """
    Represents an item in the queue waiting to be executed

    Can be either:
    - show being added (may or may not be associated with a show object)
    - show being refreshed
    - show being updated
    - show being force updated
    - show being subtitled
    """

    def __init__(self, action_id, show):
        super(ShowQueueItem, self).__init__(ShowQueueActions.names[action_id], action_id)
        self.show = show

    def is_in_queue(self) -> bool:
        return self in settings.showQueueScheduler.action.queue + [settings.showQueueScheduler.action.currentItem]

    @property
    def show_name(self) -> str:
        return self.show.name if self.show else "UNSET"

    @property
    def is_loading(self) -> bool:
        return False


class QueueItemAdd(ShowQueueItem):
    # noinspection PyPep8Naming
    def __init__(
        self,
        indexer,
        indexer_id,
        show_dir,
        default_status,
        quality,
        season_folders,
        lang,
        subtitles,
        subtitles_sc_metadata,
        anime,
        scene,
        paused,
        blacklist,
        whitelist,
        default_status_after,
        root_dir,
    ):
        super(QueueItemAdd, self).__init__(ShowQueueActions.ADD, None)

        self.show_dir = show_dir
        self.indexer = indexer
        self.indexer_id = indexer_id
        self.default_status = default_status
        self.quality = quality
        self.season_folders = season_folders
        self.lang = lang
        self.subtitles = subtitles
        self.subtitles_sc_metadata = subtitles_sc_metadata
        self.anime = anime
        self.scene = scene
        self.paused = paused
        self.blacklist = blacklist
        self.whitelist = whitelist
        self.default_status_after = default_status_after
        self.root_dir = root_dir

        self.show = None

        # Process add show in priority
        self.priority = generic_queue.QueuePriorities.HIGH

    @property
    def show_name(self):
        """
        Returns the show name if there is a show object created, if not returns
        the dir that the show is being added to.
        """
        return self.show.name if self.show else self.show_dir.rsplit(os.sep)[-1] if self.show_dir else _("Loading")

    @property
    def is_loading(self):
        """
        Returns True if we've gotten far enough to have a show object, or False
        if we still only know the folder name.
        """
        return self.show not in settings.show_list or not self.show

    @property
    def info(self):
        info = namedtuple("LoadingShowInfo", "id name sort_name network network_image_url show_image_url quality")
        if self.show:
            return info(
                id=self.show.indexerid,
                name=self.show.name,
                sort_name=self.show.sort_name,
                network=self.show.network,
                network_image_url=self.show.network_image_url,
                show_image_url=self.show.show_image_url,
                quality=self.show.quality,
            )
        # noinspection PyUnresolvedReferences
        return info(
            id=0,
            name=self.show_name,
            sort_name=sortable_name(self.show_name),
            network=_("Loading"),
            network_image_url="images/network/nonetwork.png",
            show_image_url=lambda x: "images/{}.png".format(("poster", "banner")["banner" in x]),
            quality=0,
        )

    def run(self):
        super(QueueItemAdd, self).run()

        logger.info(
            _("Starting to add show {0}").format(
                _("by show_dir: {0}").format(self.show_dir) if self.show_dir else _("by Indexer Id: {0}").format(self.indexer_id)
            )
        )
        # make sure the Indexer IDs are valid
        try:
            s = sickchill.indexer.series_by_id(indexerid=self.indexer_id, indexer=self.indexer, language=self.lang)
            if not s:
                error_string = _("Could not find show with id:{indexer_id} on {indexer}, skipping").format(
                    indexer_id=self.indexer_id, indexer=sickchill.indexer.name(self.indexer)
                )

                logger.info(error_string)
                ui.notifications.error(_("Unable to add show"), error_string)

                self._finish_early()
                return

            # Let's try to create the show Dir if it's not provided. This way we force the show dir to build using the
            # Indexers provided series name
            if self.root_dir and not self.show_dir:
                if not s.seriesName:
                    logger.info(_("Unable to get a show {show_dir}, can't add the show").format(show_dir=self.show_dir))
                    self._finish_early()
                    return

                show_dir = s.seriesName
                if settings.ADD_SHOWS_WITH_YEAR and s.firstAired:
                    try:
                        year = f"({dateutil.parser.parse(s.firstAired).year})"
                        if year not in show_dir:
                            show_dir = f"{s.seriesName} {year}"
                    except (TypeError, ValueError):
                        logger.info(_("Could not append the show year folder for the show: {show_dir}").format(show_dir=show_dir))

                self.show_dir = os.path.join(self.root_dir, sanitize_filename(show_dir))

                if settings.ADD_SHOWS_WO_DIR:
                    logger.info(_("Skipping initial creation of {show_dir} due to config.ini setting").format(show_dir=self.show_dir))
                else:
                    dir_exists = makeDir(self.show_dir)
                    if not dir_exists:
                        logger.info(_("Unable to create the folder {show_dir}, can't add the show").format(show_dir=self.show_dir))
                        self._finish_early()
                        return

                    chmodAsParent(self.show_dir)

            # this usually only happens if they have an NFO in their show dir which gave us an indexer id that has no proper english version of the show
            if getattr(s, "seriesName", None) is None:
                error_string = _(
                    "Show in {show_dir} has no name on {indexer}, probably "
                    "searched with the wrong language. Delete .nfo and add manually in the correct language."
                ).format(show_dir=self.show_dir, indexer=sickchill.indexer.name(self.indexer))

                logger.warning(error_string)
                ui.notifications.error(_("Unable to add show"), error_string)

                self._finish_early()
                return
        except Exception as error:
            error_string = (
                f"Unable to look up the show in {self.show_dir} on {sickchill.indexer.name(self.indexer)} "
                f"using ID {self.indexer_id}, not using the NFO. Delete .nfo and try adding manually again."
            )

            logger.exception(f"{error_string}: {error}")
            ui.notifications.error(_("Unable to add show"), error_string)

            if settings.USE_TRAKT:
                trakt_api = TraktAPI(settings.SSL_VERIFY, settings.TRAKT_TIMEOUT)

                title = self.show_dir.split("/")[-1]
                data = {"shows": [{"title": title, "ids": {sickchill.indexer.slug(self.indexer): self.indexer_id}}]}
                trakt_api.traktRequest("sync/watchlist/remove", data, method="POST")

            self._finish_early()
            return

        try:
            try:
                new_show = TVShow(self.indexer, self.indexer_id, self.lang)
            except MultipleShowObjectsException as error:
                # If we have the show in our list, but the location is wrong, lets fix it and refresh!
                existing_show = Show.find(settings.show_list, self.indexer_id)
                # noinspection PyProtectedMember
                if existing_show and not os.path.isdir(existing_show.get_location):
                    new_show = existing_show
                else:
                    raise error

            new_show.load_from_indexer()

            self.show = new_show

            # set up initial values
            self.show.location = self.show_dir
            self.show.subtitles = self.subtitles if self.subtitles is not None else settings.SUBTITLES_DEFAULT
            self.show.subtitles_sc_metadata = self.subtitles_sc_metadata
            self.show.quality = self.quality if self.quality else settings.QUALITY_DEFAULT
            self.show.season_folders = self.season_folders if self.season_folders is not None else settings.SEASON_FOLDERS_DEFAULT
            self.show.anime = self.anime if self.anime is not None else settings.ANIME_DEFAULT
            self.show.scene = self.scene if self.scene is not None else settings.SCENE_DEFAULT
            self.show.paused = self.paused if self.paused is not None else False

            # set up default new/missing episode status
            logger.info(_("Setting all episodes to the specified default status: {default_status}").format(default_status=self.default_status))
            self.show.default_ep_status = self.default_status

            if self.show.anime:
                self.show.release_groups = BlackAndWhiteList(self.show.indexerid)
                if self.blacklist:
                    self.show.release_groups.set_black_keywords(self.blacklist)
                if self.whitelist:
                    self.show.release_groups.set_white_keywords(self.whitelist)

            # # be smart-ish about this
            # if self.show.genre and 'talk show' in self.show.genre.lower():
            #     self.show.air_by_date = 1
            # if self.show.genre and 'documentary' in self.show.genre.lower():
            #     self.show.air_by_date = 0
            # if self.show.classification and 'sports' in self.show.classification.lower():
            #     self.show.sports = 1

        except Exception as error:
            error_string = f"Unable to add {self.show.name if self.show else 'show'} due to an error with {sickchill.indexer.name(self.indexer)}"

            logger.exception(f"{error_string}: {error}")

            logger.exception(f"Error trying to add show: {error}")
            logger.debug(traceback.format_exc())

            ui.notifications.error(_("Unable to add show"), error_string)

            self._finish_early()
            return

        except MultipleShowObjectsException:
            error_string = _("The show in {show_dir} is already in your show list, skipping").format(show_dir=self.show_dir)
            logger.warning(error_string)
            ui.notifications.error(_("Show skipped"), error_string)

            self._finish_early()
            return

        self.show.load_imdb_info()

        try:
            self.show.save_to_db()
        except Exception as error:
            logger.exception(f"Error saving the show to the database: {error}")
            logger.debug(traceback.format_exc())
            self._finish_early()
            raise

        # add it to the show list
        if not Show.find(settings.show_list, self.indexer_id):
            settings.show_list.append(self.show)

        # Presort show_list, so we don't have to do it every page load
        settings.show_list = sorted(settings.show_list, key=attrgetter("sort_name"))

        try:
            self.show.load_episodes_from_indexer(force_all=True)
        except Exception as error:
            logger.exception(f"Error with {self.show.idxr.name}, not creating episode list: {error}")
            logger.debug(traceback.format_exc())

        # update internal name cache
        name_cache.build_name_cache(self.show)

        try:
            self.show.load_episodes_from_dir()
        except Exception as error:
            logger.exception(f"Error searching dir for episodes: {error}")
            logger.debug(traceback.format_exc())

        # if they set default ep status to WANTED then run the backlog to search for episodes
        # FIXME: This needs to be a backlog queue item!!!
        if self.show.default_ep_status == WANTED:
            logger.info("Launching backlog for this show since its episodes are WANTED")
            settings.backlogSearchScheduler.action.searchBacklog([self.show])

        self.show.write_metadata()
        self.show.update_metadata()
        self.show.populate_cache()

        self.show.flush_episodes()

        if settings.USE_TRAKT:
            # if there are specific episodes that need to be added by trakt
            settings.traktCheckerScheduler.action.check_new_show(self.show)
            # add show to trakt.tv library
            if settings.TRAKT_SYNC:
                settings.traktCheckerScheduler.action.add_show_to_trakt_library(self.show)

            if settings.TRAKT_SYNC_WATCHLIST:
                logger.info("update watchlist")
                notifiers.trakt_notifier.update_watchlist(show_obj=self.show)

        # Load XEM data to DB for show
        scene_numbering.xem_refresh(self.show.indexerid, self.show.indexer, force=True)

        # check if show has XEM mapping, so we can determine if searches should go by scene numbering or indexer numbering.
        if not self.scene and scene_numbering.get_xem_numbering_for_show(self.show.indexerid, self.show.indexer):
            self.show.scene = 1

        # After initial add, set to default_status_after.
        self.show.default_ep_status = self.default_status_after

        super(QueueItemAdd, self).finish()
        self.finish()

    def _finish_early(self):
        if self.show is not None:
            Show.delete(self.show)

        super(QueueItemAdd, self).finish()
        self.finish()


class QueueItemRefresh(ShowQueueItem):
    def __init__(self, show: "TVShow" = None, force=False):
        super(QueueItemRefresh, self).__init__(ShowQueueActions.REFRESH, show)

        # do refreshes first because they're quick
        self.priority = generic_queue.QueuePriorities.HIGH

        # force refresh certain items
        self.force = force

    def run(self):
        super(QueueItemRefresh, self).run()

        logger.info(f"Performing refresh on {self.show.name}")

        self.show.refresh_dir()
        self.show.write_metadata()
        if self.force:
            self.show.update_metadata()
        self.show.populate_cache()

        # Load XEM data to DB for show
        scene_numbering.xem_refresh(self.show.indexerid, self.show.indexer)

        super(QueueItemRefresh, self).finish()
        self.finish()


class QueueItemRename(ShowQueueItem):
    def __init__(self, show: "TVShow" = None):
        super(QueueItemRename, self).__init__(ShowQueueActions.RENAME, show)

    def run(self):
        super(QueueItemRename, self).run()

        logger.info(f"Performing rename on {self.show.name}")

        try:
            self.show.location
        except ShowDirectoryNotFoundException:
            logger.warning(f"Can't perform rename on {self.show.name} when the show dir is missing.")
            super(QueueItemRename, self).finish()
            self.finish()
            return

        ep_obj_rename_list = []

        ep_obj_list = self.show.get_all_episodes(has_location=True)
        for cur_ep_obj in ep_obj_list:
            # Only want to rename if we have a location
            if cur_ep_obj.location:
                # do we have one of multi-episodes in the rename list already
                for cur_related_ep in cur_ep_obj.related_episodes + [cur_ep_obj]:
                    if cur_related_ep in ep_obj_rename_list:
                        break
                else:
                    ep_obj_rename_list.append(cur_ep_obj)

        for cur_ep_obj in ep_obj_rename_list:
            cur_ep_obj.rename()

        super(QueueItemRename, self).finish()
        self.finish()


class QueueItemSubtitle(ShowQueueItem):
    def __init__(self, show: "TVShow" = None):
        super(QueueItemSubtitle, self).__init__(ShowQueueActions.SUBTITLE, show)

    def run(self):
        super(QueueItemSubtitle, self).run()

        logger.info(f"Downloading subtitles for {self.show.name} ")

        self.show.download_subtitles()

        super(QueueItemSubtitle, self).finish()
        self.finish()


class QueueItemUpdate(ShowQueueItem):
    def __init__(self, show: "TVShow" = None, force=False):
        action = ShowQueueActions.FORCEUPDATE if force else ShowQueueActions.UPDATE
        super(QueueItemUpdate, self).__init__(action, show)
        self.force = force
        self.priority = generic_queue.QueuePriorities.HIGH

    def run(self):
        super(QueueItemUpdate, self).run()

        logger.debug(f"Beginning update of {self.show.name}")

        logger.debug(f"Retrieving show info from {self.show.idxr.name}")
        try:
            self.show.load_from_indexer()
        except Exception as error:
            logger.warning(f"Unable to contact {self.show.idxr.name}, aborting: {error}")
            super(QueueItemUpdate, self).finish()
            self.finish()
            return

        self.show.load_imdb_info()

        # have to save show before reading episodes from db
        try:
            self.show.save_to_db()
        except Exception as error:
            logger.exception(f"Error saving show info to the database: {error}")
            logger.debug(traceback.format_exc())

        # get episode list from DB
        database_episodes = self.show.load_episodes_from_db()

        # get episode list from TVDB
        logger.debug(f"Loading all episodes from {self.show.idxr.name}")
        try:
            indexer_episodes = self.show.load_episodes_from_indexer(self.force)
        except Exception as error:
            logger.exception(f"Unable to get info from {self.show.idxr.name}, the show info will not be refreshed: {error}")
            indexer_episodes = dict()

        for season in indexer_episodes:
            for episode in indexer_episodes[season]:
                self.show.get_episode(season, episode).save_to_db()
                if season in database_episodes and episode in database_episodes[season]:
                    del database_episodes[season][episode]

        if indexer_episodes:
            # remaining episodes in the database list are not on the indexer, just delete them from the DB
            # TODO: database: maybe add a "marked for deletion" column to the database, and when we delete an episode just mark it and do the actual deletion after a few days.
            #  episodes marked for deletion can be hidden from the ui, but we wont lose downloaded files for episodes that are then brought back and they wont be sent to download again?
            for season in database_episodes:
                for episode in database_episodes[season]:
                    logger.info("Permanently deleting episode {0:02d}E{1:02d} from the database".format(season, episode))
                    try:
                        self.show.get_episode(season, episode).delete_episode()
                    except EpisodeDeletedException:
                        pass

        #  save show again, in case episodes have changed
        try:
            self.show.save_to_db()
        except Exception as error:
            logger.exception(f"Error saving show info to the database: {error}")
            logger.debug(traceback.format_exc())

        logger.debug(f"Finished update of {self.show.name}")

        # oldbeard.showQueueScheduler.action.refresh_show(self.show, self.force)
        QueueItemRefresh(self.show, self.force).run()
        super(QueueItemUpdate, self).finish()
        self.finish()


class QueueItemRemove(ShowQueueItem):
    def __init__(self, show: "TVShow" = None, full: bool = False):
        super(QueueItemRemove, self).__init__(ShowQueueActions.REMOVE, show)

        # noinspection IncorrectFormatting
        self.priority = generic_queue.QueuePriorities.HIGH**2
        self.full = full

    def run(self):
        super(QueueItemRemove, self).run()
        logger.info(f"Removing {self.show.name}")
        self.show.delete_show(full=self.full)

        if settings.USE_TRAKT:
            try:
                settings.traktCheckerScheduler.action.remove_show_from_trakt_library(self.show)
            except Exception as error:
                logger.warning(_("Unable to delete show from Trakt: {show_name}. Error: {error}").format(show_name=self.show.name, error=error))

        # If any notification fails, don't stop removal
        # noinspection PyBroadException
        try:
            # TODO: episode_object is undefined here, so all of these will fail.
            # send notifications
            # notifiers.notify_download(episode_object.format_pattern('%SN - %Sx%0E - %EN - %QN'))

            # do the library update for KODI
            notifiers.kodi_notifier.update_library(self.show.name)

            # do the library update for Plex
            notifiers.plex_notifier.update_library(self.show)

            # do the library update for EMBY
            notifiers.emby_notifier.update_library(self.show)

            # do the library update for NMJ
            # nmj_notifier kicks off its library update when the notify_download is issued (inside notifiers)

            # do the library update for Synology Indexer
            notifiers.synoindex_notifier.addFolder(self.show.get_location)

            # do the library update for tivo
            notifiers.pytivo_notifier.update_library(self.show)
        except Exception:
            logger.info(_("Some notifications could not be sent. Continuing removal of {}...").format(self.show.name))

        super(QueueItemRemove, self).finish()
        self.finish()
